import math
import os
import sys
from random import random
import re
from threading import Thread
from time import perf_counter
import time
from typing import Optional
import uuid

import modal


app = modal.App("vita_audio")
vita_audio_img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "clang", "cmake", "ninja-build")
    .run_commands(
        "git clone -b feat/achatbot https://github.com/weedge/VITA-Audio.git",
        "cd /VITA-Audio && git submodule update --init --recursive",
        "cd /VITA-Audio && pip install -q -r requirements_ds_gpu.txt",
    )
    .pip_install(
        "torch==2.6.0",
        "torchaudio==2.6.0",
        "torchvision==0.21.0",
        index_url="https://download.pytorch.org/whl/cu126",
    )
    .pip_install("wheel")
    # https://github.com/Dao-AILab/flash-attention/releases/tag/v2.7.4.post1
    .pip_install("flash-attn", extra_options="--no-build-isolation")
    .run_commands(
        "cd /VITA-Audio && git pull origin feat/achatbot",
        "cd /VITA-Audio && git checkout 32fb8180fcd7a4bbaa8c9dd77928c78cb7fd26ce",
    )
    .run_commands(
        "git clone https://github.com/weedge/CosyVoice.git",
        "cd /CosyVoice && git checkout 6507762ba72ecfd4f9c15e17f543c8aaeef2fd8f",
        "pip install rich",
    )
    .pip_install("snac")
    .run_commands(
        "git clone https://github.com/weedge/Spark-TTS.git",
        "cd /Spark-TTS && git checkout 244144322c738a64d98ce51247d8de75a74e9449",
    )
    .pip_install("einops", "einx", "omegaconf", "packaging", "safetensors", "soxr")
    .env(
        {
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
            "TQDM_DISABLE": "1",
            "AUDIO_TOKENIZER_TYPE": os.getenv("AUDIO_TOKENIZER_TYPE", "sensevoice_glm4voice"),
            # VITA-MLLM/VITA-Audio-Boost | VITA-MLLM/VITA-Audio-Balance
            "MTP_LLM_MODEL": os.getenv("MTP_LLM_MODEL", "VITA-MLLM/VITA-Audio-Balance"),
        }
    )
)

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)

with vita_audio_img.imports():
    import subprocess
    import torch
    import numpy as np
    import tqdm
    from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer, TextIteratorStreamer
    from transformers.generation import GenerationConfig

    sys.path.append("/VITA-Audio")
    sys.path.append("/VITA-Audio/third_party/GLM-4-Voice/")
    sys.path.append("/VITA-Audio/third_party/GLM-4-Voice/cosyvoice/")
    sys.path.append("/VITA-Audio/third_party/GLM-4-Voice/third_party/Matcha-TTS/")

    from vita_audio.tokenizer import get_audio_tokenizer
    from vita_audio.data.processor.audio_processor import add_audio_input_contiguous

    # audio vq codec model | cosyvoice (glm-4-voice-tokenizer or FunAudioLLM/CosyVoice2-0.5B tokenizer + vocoder)
    audio_tokenizer_model_path = None
    sense_voice_model_path = None  # sensevoice_small model
    flow_path = None
    model_path = None

    # glm4voice | cosyvoice2 | snac24khz | sensevoice_sparktts | sensevoice_glm4voice
    support_types = [
        "glm4voice",
        "sensevoice_glm4voice",
        # "snac24khz",
        # "cosyvoice2",
        # "sensevoice_sparktts",
    ]
    audio_tokenizer_type = os.getenv("AUDIO_TOKENIZER_TYPE", "sensevoice_glm4voice")
    print(f"{audio_tokenizer_type=}")
    mtp_llm_model = os.path.join(
        HF_MODEL_DIR, os.getenv("MTP_LLM_MODEL", "VITA-MLLM/VITA-Audio-Balance")
    )
    print(f"{mtp_llm_model=}")

    if audio_tokenizer_type == "sensevoice_glm4voice":
        # sensevoice_glm4voice
        # (sensevoice_small WaveFrontend extract audio feature)
        sense_voice_model_path = os.path.join(HF_MODEL_DIR, "FunAudioLLM/SenseVoiceSmall")
        # glm-4-voice-decoder (flow + hift) from cosyvoice (support streaming)
        flow_path = os.path.join(HF_MODEL_DIR, "THUDM/glm-4-voice-decoder")
        # sensevoice_small encoder(no ctc head) + qwen2 model (no mtp)
        # model_path = os.path.join(HF_MODEL_DIR, "VITA-MLLM/VITA-Audio-Plus-Vanilla")
        model_path = mtp_llm_model

    if audio_tokenizer_type == "glm4voice":
        audio_tokenizer_model_path = os.path.join(HF_MODEL_DIR, "THUDM/glm-4-voice-tokenizer")
        flow_path = os.path.join(HF_MODEL_DIR, "THUDM/glm-4-voice-decoder")

        model_path = mtp_llm_model
        # run is ok, but don't use this model
        # model_path = os.path.join(HF_MODEL_DIR, "VITA-MLLM/VITA-Audio-Plus-Vanilla")

    if audio_tokenizer_type == "cosyvoice2":
        sys.path.insert(0, "/CosyVoice")
        # (sensevoice_small WaveFrontend extract audio feature)
        audio_tokenizer_model_path = os.path.join(HF_MODEL_DIR, "FunAudioLLM/CosyVoice2-0.5B")

        model_path = mtp_llm_model

        # run is ok, but don't use this model
        # model_path = os.path.join(HF_MODEL_DIR, "VITA-MLLM/VITA-Audio-Plus-Vanilla")

    if audio_tokenizer_type == "snac24khz":
        audio_tokenizer_model_path = os.path.join(HF_MODEL_DIR, "hubertsiuzdak/snac_24khz")

        model_path = mtp_llm_model

        # run is ok, but don't use this model
        # model_path = os.path.join(HF_MODEL_DIR, "VITA-MLLM/VITA-Audio-Plus-Vanilla")

    if audio_tokenizer_type == "sensevoice_sparktts":
        sys.path.insert(0, "/Spark-TTS")
        # bicodec (vq codec)
        audio_tokenizer_model_path = os.path.join(HF_MODEL_DIR, "SparkAudio/Spark-TTS-0.5B")

        model_path = mtp_llm_model

        # run is ok, but don't use this model
        # sense_voice_model_path = os.path.join(HF_MODEL_DIR, "FunAudioLLM/SenseVoiceSmall")
        # model_path = os.path.join(HF_MODEL_DIR, "VITA-MLLM/VITA-Audio-Plus-Vanilla")


@app.function(
    gpu=os.getenv("IMAGE_GPU", None),
    cpu=2.0,
    retries=1,
    image=vita_audio_img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        ASSETS_DIR: assets_dir,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
def run(func):
    if audio_tokenizer_type not in support_types:
        print(
            f"Invalid audio_tokenizer_type: {audio_tokenizer_type}, please choose one of the following: {support_types}"
        )
        return
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    gpu_prop = torch.cuda.get_device_properties("cuda")
    print(gpu_prop)

    func()


def tokenize():
    from evaluation.get_chat_template import qwen2_chat_template as chat_template

    config = AutoConfig.from_pretrained(
        model_path,
        trust_remote_code=True,
    )
    print(f"{config=}")

    tokenizer = AutoTokenizer.from_pretrained(
        model_path,
        trust_remote_code=True,
        chat_template=chat_template,
    )
    # print(f"{tokenizer=}")
    print(f"{tokenizer.get_chat_template()=}")


def dump_model():
    # https://huggingface.co/VITA-MLLM/VITA-Audio-Boost/blob/main/modeling_qwen2.py#L781
    # https://huggingface.co/VITA-MLLM/VITA-Audio-Balance/blob/main/modeling_qwen2.py#L781
    # https://huggingface.co/VITA-MLLM/VITA-Audio-Plus-Vanilla/blob/main/modeling_qwen2.py#L834
    for name, model_path in {
        "Boost": os.path.join(HF_MODEL_DIR, "VITA-MLLM/VITA-Audio-Boost"),
        "Balance": os.path.join(HF_MODEL_DIR, "VITA-MLLM/VITA-Audio-Balance"),
        "Vanilla": os.path.join(HF_MODEL_DIR, "VITA-MLLM/VITA-Audio-Plus-Vanilla"),
    }.items():
        model = AutoModelForCausalLM.from_pretrained(
            model_path,
            trust_remote_code=True,
            device_map="cuda:0",
            torch_dtype=torch.bfloat16,
            attn_implementation="flash_attention_2",
        ).eval()
        print_model_params(model, f"Qwen2MTP_AR_LLM_{name}")
        del model
        torch.cuda.empty_cache()

    audio_tokenizer = get_audio_tokenizer(
        model_type=audio_tokenizer_type,
        flow_path=flow_path,
        rank=0,
        audio_tokenizer_model_path=audio_tokenizer_model_path,
        sense_voice_model_path=sense_voice_model_path,
    )
    print(f"{audio_tokenizer=}")

    audio_tokenizer.load_model()
    print(f"SenseVoiceSmall args: {audio_tokenizer.kwargs}")
    print_model_params(audio_tokenizer.sensevoice_model, "audio_tokenizer.audio_encoder.sensevoice")
    print_model_params(audio_tokenizer.whisper_model, "audio_tokenizer.audio_encoder.whisper")

    print_model_params(audio_tokenizer.audio_decoder.flow, "audio_tokenizer.audio_decoder.flow")
    print_model_params(audio_tokenizer.audio_decoder.hift, "audio_tokenizer.audio_decoder.hift")


def benchmark_llm():
    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        audio_tokenizer_model_path=audio_tokenizer_model_path,
        sense_voice_model_path=sense_voice_model_path,
        flow_path=flow_path,
    )
    for mtp_inference_mode, tag in zip(
        [
            [8192, 0],
            [1, 4, 3, 8, 4, 10],
            [1, 10, 4, 10],
            [1, 10],
        ],
        [
            "Vanilla",
            "Balance",
            "Boost",
            "Turbo",
        ],
    ):
        print("=" * 100)
        print("benchmark_llm")
        print(f"{tag}")

        s2s_inference.benchmark_forward(mtp_inference_mode)

        s2s_inference.benchmark_generate(mtp_inference_mode)

        generated_text = ""
        for new_text in s2s_inference.benchmark_generate_stream(
            mtp_inference_mode=mtp_inference_mode
        ):
            generated_text += new_text
            # print(new_text, end="", flush=True)


def benchmark_sts():
    import torchaudio

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        audio_tokenizer_model_path=audio_tokenizer_model_path,
        sense_voice_model_path=sense_voice_model_path,
        flow_path=flow_path,
    )

    audio_paths = [
        "/VITA-Audio/asset/介绍一下上海.wav",
        "/VITA-Audio/asset/发表一个悲伤的演讲.wav",
        "/VITA-Audio/asset/发表一个振奋人心的演讲.wav",
    ]
    output_dir = ASSETS_DIR

    for _ in range(10):
        print("=" * 100)
        print("benchmark_sts")
        audio_path = random.choice(audio_paths)
        print(f"{audio_path}")

        start = time.time()
        audio_idx = 0
        generated_text = ""
        all_tts_speech = []
        past_tts_speech_len = 0
        for new_text in s2s_inference.run_infer_stream(audio_path=audio_path):
            # print(new_text, end="", flush=True)

            generated_text += new_text

            if new_text == "<|end_of_audio|>":
                audio_tokens = extract_token_ids_as_int(generated_text)

                tts_speech = s2s_inference.audio_tokenizer.decode(audio_tokens, option_steps=1)
                tts_speech = tts_speech[past_tts_speech_len:]
                past_tts_speech_len += len(tts_speech)
                all_tts_speech.append(tts_speech)

                end = time.time()
                if audio_idx == 0:
                    print(audio_tokens)
                print(f"{audio_idx} audio chunk {end - start}")

                wav_path = os.path.join(output_dir, audio_path[:-4] + f"_{audio_idx}.wav")
                os.makedirs(os.path.dirname(wav_path), exist_ok=True)
                torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")

                audio_idx += 1
                start = time.time()

        wav_path = os.path.join(output_dir, audio_path[:-4] + ".wav")
        tts_speech = torch.cat(all_tts_speech, dim=0)
        os.makedirs(os.path.dirname(wav_path), exist_ok=True)
        torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")


# ==============================================================
# Text
def text():
    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        sense_voice_model_path=None,
        flow_path=None,
    )

    for text in [
        # "How many helicopters can a human eat in one sitting?",
        # "你叫什么名字？",
        # "写一首诗",
        "介绍一下上海",
    ]:
        print("=" * 100)
        print("text_task")
        print(f"{text=}")

        start = perf_counter()
        output, _ = s2s_inference.run_infer(
            message=text,
            mode=None,
            # do_sample=True,
            mtp_inference_mode=[8192, 0],
        )
        print(f"{output=}", flush=True)
        print(f"cost: {perf_counter()-start} s")


def text_audio():
    import torchaudio

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        sense_voice_model_path=None,
        flow_path=flow_path,
    )

    for text in [
        "How many helicopters can a human eat in one sitting?",
        "你叫什么名字？",
        "写一首诗",
        "介绍一下北京",
    ]:
        print("=" * 100)
        print("text_audio_task")
        print(f"{text=}")

        start = perf_counter()
        output, tts_speech = s2s_inference.run_infer(
            message=text,
            mode="luke",
            do_sample=False,
            # mtp_inference_mode=[8192, 0],
        )
        print(f"{output=}", flush=True)
        print(f"{tts_speech.shape=}", flush=True)
        print(f"cost: {perf_counter()-start} s")

        file_name = text[:16]
        wav_path = os.path.join(ASSETS_DIR, f"sts_{file_name}.wav")
        print(f"save to {wav_path=}")
        os.makedirs(os.path.dirname(wav_path), exist_ok=True)
        torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")


# ==============================================================
# Text stream
def text_stream():
    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        sense_voice_model_path=None,
        flow_path=None,
    )
    for text in [
        "你叫什么名字？",
        "讲一个儿童故事",
    ]:
        print("=" * 100)
        print("text_stream_task")
        print(f"{text=}")

        times = []
        start_time = perf_counter()
        generated_text = ""
        for new_text in s2s_inference.run_infer_stream(
            message=text,
            mode=None,
            # do_sample=True,
            mtp_inference_mode=[8192, 0],
        ):
            times.append(perf_counter() - start_time)
            generated_text += new_text
            print(new_text, end="", flush=True)
            start_time = perf_counter()

        print(
            f"\ngenerate [{generated_text}] first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s\n"
        )


def text_audio_stream():
    import soundfile as sf

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        sense_voice_model_path=None,
        flow_path=flow_path,
    )
    for text in [
        "你叫什么名字？",
        # "请讲一个儿童故事。",
    ]:
        print("=" * 100)
        print("text_audio_stream_task")
        print(f"{text=}")

        times = []
        start_time = perf_counter()
        generated_text = ""
        for new_text in s2s_inference.run_infer_stream(
            message=text,
            mode="luke",
            do_sample=True,
            # mtp_inference_mode=[8192, 0],
        ):
            times.append(perf_counter() - start_time)
            generated_text += new_text
            print(new_text, end=" ", flush=True)
            start_time = perf_counter()

        print(
            f"\ngenerate [{generated_text}] first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s\n"
        )
        file_name = text[:16]
        audio_decode_time = []
        audios = []
        audio_segments = find_audio_segments_regex(generated_text)
        for audio_idx, audio_segment in enumerate(audio_segments):
            start = perf_counter()

            audio_tokens = extract_token_ids_as_int(audio_segment)
            print(f"{audio_tokens=}")

            tts_speech = s2s_inference.audio_tokenizer.decode(audio_tokens)
            audio_decode_time.append(perf_counter() - start)

            # wav_path = os.path.join(ASSETS_DIR, f"qa_stream_{audio_idx}_{file_name}.wav")
            # print(f"save to {wav_path=}")
            # torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")

            audios.append(tts_speech.cpu().numpy())

        wav_path = os.path.join(ASSETS_DIR, f"sts_stream_{file_name}.wav")
        print(f"save to {wav_path=}")
        sf.write(wav_path, np.concatenate(audios), samplerate=22050)
        info = sf.info(wav_path, verbose=True)
        print(
            f"\ngenerate first audio segment cost time: {audio_decode_time[0]} s, {len(audio_decode_time)} segment cost time: {sum(audio_decode_time)} s | wav duration: {info.duration} s | audio tokenizer decode RTF: {sum(audio_decode_time)/info.duration} \n"
        )


# ==============================================================
# S2S
def sts():
    import torchaudio

    output_dir = ASSETS_DIR
    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        audio_tokenizer_model_path=audio_tokenizer_model_path,
        sense_voice_model_path=sense_voice_model_path,
        flow_path=flow_path,
    )

    for audio_path in [
        "/VITA-Audio/asset/介绍一下上海.wav",
        # "/VITA-Audio/asset/发表一个悲伤的演讲.wav",
        # "/VITA-Audio/asset/发表一个振奋人心的演讲.wav",
        # "/VITA-Audio/asset/piano.mp3",
    ]:
        print("=" * 100)
        print("sts_task")
        print(f"{audio_path=}")
        subprocess.run(f"cp {audio_path} {ASSETS_DIR}", shell=True)

        start_time = perf_counter()
        output, tts_speech = s2s_inference.run_infer(
            audio_path=audio_path,
        )
        print(f"{output=}")
        print(f"{tts_speech.shape=}", flush=True)
        print(f"cost time: {perf_counter()-start_time:.6f} s")

        file_name = os.path.basename(audio_path)
        wav_path = os.path.join(output_dir, f"sts_{file_name}")
        print(f"save to {wav_path=}")
        os.makedirs(os.path.dirname(wav_path), exist_ok=True)
        torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")


# ==============================================================
# S2S stream
def sts_stream():
    """
    audio decode need look-ahead
    """
    import torchaudio
    import soundfile as sf

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        audio_tokenizer_model_path=audio_tokenizer_model_path,
        sense_voice_model_path=sense_voice_model_path,
        flow_path=flow_path,
    )

    for audio_path in [
        "/VITA-Audio/asset/介绍一下上海.wav",  # warmup
        "/VITA-Audio/asset/介绍一下上海.wav",
    ]:
        print("=" * 100)
        print("sts_stream_task")
        print(f"{audio_path=}")

        times = []
        start_time = perf_counter()
        generated_text = ""
        for new_text in s2s_inference.run_infer_stream(audio_path=audio_path):
            times.append(perf_counter() - start_time)
            generated_text += new_text
            print(new_text, end="")
            start_time = perf_counter()
        print(
            f"\ngenerate [{generated_text}] first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s\n"
        )

        file_name = os.path.basename(audio_path)
        audio_decode_time = []
        audios = []
        audio_segments = find_audio_segments_regex(generated_text)
        for audio_idx, audio_segment in enumerate(audio_segments):
            start = perf_counter()

            audio_tokens = extract_token_ids_as_int(audio_segment)
            print(f"{audio_tokens=}")

            tts_speech = s2s_inference.audio_tokenizer.decode(audio_tokens)
            audio_decode_time.append(perf_counter() - start)

            # wav_path = os.path.join(ASSETS_DIR, f"qa_stream_{audio_idx}_{file_name}")
            # print(f"save to {wav_path=}")
            # torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")

            audios.append(tts_speech.cpu().numpy())

        wav_path = os.path.join(ASSETS_DIR, f"sts_stream_{file_name}")
        print(f"save to {wav_path=}")
        sf.write(wav_path, np.concatenate(audios), samplerate=22050)
        info = sf.info(wav_path, verbose=True)

        print(
            f"\ngenerate first audio segment cost time: {audio_decode_time[0]} s, {len(audio_decode_time)} segment cost time: {sum(audio_decode_time)} s | wav duration: {info.duration} s | audio tokenizer decode RTF: {sum(audio_decode_time)/info.duration} \n"
        )


# ==============================================================
# ASR
def asr():
    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        audio_tokenizer_model_path=audio_tokenizer_model_path,
        sense_voice_model_path=sense_voice_model_path,
        flow_path=None,
    )

    for audio_path in [
        "/VITA-Audio/asset/介绍一下上海.wav",
        "/VITA-Audio/asset/发表一个悲伤的演讲.wav",
        "/VITA-Audio/asset/发表一个振奋人心的演讲.wav",
    ]:
        print("=" * 100)
        print("asr_task")
        print(f"{audio_path=}")

        start_time = perf_counter()
        output, _ = s2s_inference.run_infer(
            audio_path=audio_path,
            # message="Translate the speech to english.",
            # message="Translate the speech to text.",
            message="Convert the speech to text.",
            mode=None,
        )
        print(f"{output=}", flush=True)
        print(f"cost: {perf_counter() - start_time}")


def asr_text_stream():
    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        audio_tokenizer_model_path=audio_tokenizer_model_path,
        sense_voice_model_path=sense_voice_model_path,
        flow_path=None,
    )

    for audio_path in [
        # os.path.join(ASSETS_DIR, "asr_example_zh.wav"),
        "/VITA-Audio/asset/介绍一下上海.wav",
        # "/VITA-Audio/asset/发表一个悲伤的演讲.wav",
        # "/VITA-Audio/asset/发表一个振奋人心的演讲.wav",
    ]:
        print("=" * 100)
        print("asr_text_stream_task")
        print(f"{audio_path=}")

        times = []
        start_time = perf_counter()
        generated_text = ""
        for new_text in s2s_inference.run_infer_stream(
            audio_path=audio_path,
            # message="Translate the speech to text.",
            message="Convert the speech to text.",
            mode=None,
        ):
            times.append(perf_counter() - start_time)
            generated_text += new_text
            print(new_text, end=" ", flush=True)
            start_time = perf_counter()
        print(
            f"\ngenerate [{generated_text}] first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s\n"
        )


def asr_live():
    """
    need ft to support speech streaming asr
    see: https://github.com/FunAudioLLM/SenseVoice small
    """
    pass


# ==============================================================
# TTS
def tts():
    import torchaudio

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        flow_path=flow_path,
    )

    # bad case: use glm4voice tokenizer to generate tts with Boost/Balance mtp llm(7B)
    TTS_texts = [
        # "我们将为全球城市的可持续发展贡献力量。",
        # "通天河 灵感大王",
        # "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",
        # "一二三四五六七八九十",
        # "One Two Tree Four Five Six Seven Eight Night Ten",
        # "1 2 3 4 5 6 7 8 9 10",
        # "12345678910",
        # "两个黄鹂鸣翠柳，一行白鹭上青天。窗含西岭千秋雪，门泊东吴万里船。",
        # "坡上立着一只鹅，坡下就是一条河。宽宽的河，肥肥的鹅，鹅要过河，河要渡鹅不知是鹅过河，还是河渡鹅?",
        # "扁担长，板凳宽，扁担没有板凳宽，板凳没有扁担长。扁担绑在板凳上，板凳不让扁担绑在板凳上。",
        # "化肥会挥发，黑化肥发灰，灰化肥发黑。黑化肥发灰会挥发；灰化肥挥发会发黑。黑化肥挥发发灰会花飞；灰化肥挥发发黑会飞花，黑灰化肥会挥发发灰黑讳为花飞；灰黑化肥会挥发发黑灰为讳飞花。",
        # "圆桌儿、方桌儿没有腿儿，墨水瓶儿里没有水儿，花瓶里有花儿没有叶儿，练习本儿上写字儿没有准儿，甘蔗好吃净是节儿。西瓜挺大没有味儿，坛儿里的小米儿长了虫儿，鸡毛掸子成了棍儿，水缸沿儿上系围裙儿，耗子打更猫打盹儿，新买的小褂儿没钉扣儿，奶奶想说没有劲儿。",
        "起床歌：小宝宝，起得早，睁开眼，眯眯笑，咿呀呀，学说话，伸伸手，要人抱。穿衣歌小胳膊，穿袖子，穿上衣，扣扣子，小脚丫，穿裤子，穿上袜子穿鞋子。小镜子-小镜子，圆又圆，看宝宝，露笑脸。闭上眼，做个梦，变月亮，挂上天。小铃铛叮铃铃，叮铃铃，一会远，一会近。小宝宝，耳朵灵，听铃声，找到铃。学画画小宝宝，学画画，大蜡笔，手中拿，画小鸭，叫嘎嘎，画小马，骑回家。大鞋子大鞋子，像只船，爸爸穿，我也穿，一二一，向前走，走呀走，翻了船。逛公园逛公园，宝宝笑，东看看，西瞧瞧，花儿香，鸟儿叫，小草绿，小树摇。看画报小娃娃，看画报，睁大眼，仔细瞧，布娃娃，哈哈笑，伸伸手，要你抱。搭积木大积木，红黄兰，小宝宝，最爱玩，搭火车，钻山洞，盖高楼，连着天。小汽车小汽车，嘀嘀嘀，开过来，开过去，小宝宝，当司机，送妈妈，上班去。藏猫猫儿歌：躲猫猫，躲猫猫， 猫猫、猫猫在哪里？喵……猫咪在这里。",
    ]

    for text in TTS_texts:
        print("=" * 100)
        print("tts_task")
        print(f"{text=}")

        start = perf_counter()
        output, tts_speech = s2s_inference.run_infer(
            message="Convert the text to speech.\n" + text,
            mode=None,
            do_sample=True,
        )
        print(f"{output=}", flush=True)
        print(f"{tts_speech.shape=}", flush=True)
        print(f"cost: {perf_counter()-start}")

        wav_path = os.path.join(ASSETS_DIR, text[:16] + ".wav")
        print(f"save to {wav_path}")
        torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")


# ==============================================================
# Clone TTS
def tts_clone():
    import torchaudio

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        audio_tokenizer_model_path=audio_tokenizer_model_path,
        sense_voice_model_path=sense_voice_model_path,
        flow_path=flow_path,
    )

    TTS_texts = [
        "我们将为全球城市的可持续发展贡献力量。",
        # "通天河 灵感大王",
        # "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",
        # "一二三四五六七八九十",
        # "One Two Tree Four Five Six Seven Eight Night Ten",
        # "1 2 3 4 5 6 7 8 9 10",
        # "12345678910",
        # "两个黄鹂鸣翠柳，一行白鹭上青天。窗含西岭千秋雪，门泊东吴万里船。",
        # "坡上立着一只鹅，坡下就是一条河。宽宽的河，肥肥的鹅，鹅要过河，河要渡鹅不知是鹅过河，还是河渡鹅?",
        # "扁担长，板凳宽，扁担没有板凳宽，板凳没有扁担长。扁担绑在板凳上，板凳不让扁担绑在板凳上。",
        # "化肥会挥发，黑化肥发灰，灰化肥发黑。黑化肥发灰会挥发；灰化肥挥发会发黑。黑化肥挥发发灰会花飞；灰化肥挥发发黑会飞花，黑灰化肥会挥发发灰黑讳为花飞；灰黑化肥会挥发发黑灰为讳飞花。",
        # "圆桌儿、方桌儿没有腿儿，墨水瓶儿里没有水儿，花瓶里有花儿没有叶儿，练习本儿上写字儿没有准儿，甘蔗好吃净是节儿。西瓜挺大没有味儿，坛儿里的小米儿长了虫儿，鸡毛掸子成了棍儿，水缸沿儿上系围裙儿，耗子打更猫打盹儿，新买的小褂儿没钉扣儿，奶奶想说没有劲儿。",
        # "起床歌：小宝宝，起得早，睁开眼，眯眯笑，咿呀呀，学说话，伸伸手，要人抱。穿衣歌小胳膊，穿袖子，穿上衣，扣扣子，小脚丫，穿裤子，穿上袜子穿鞋子。小镜子-小镜子，圆又圆，看宝宝，露笑脸。闭上眼，做个梦，变月亮，挂上天。小铃铛叮铃铃，叮铃铃，一会远，一会近。小宝宝，耳朵灵，听铃声，找到铃。学画画小宝宝，学画画，大蜡笔，手中拿，画小鸭，叫嘎嘎，画小马，骑回家。大鞋子大鞋子，像只船，爸爸穿，我也穿，一二一，向前走，走呀走，翻了船。逛公园逛公园，宝宝笑，东看看，西瞧瞧，花儿香，鸟儿叫，小草绿，小树摇。看画报小娃娃，看画报，睁大眼，仔细瞧，布娃娃，哈哈笑，伸伸手，要你抱。搭积木大积木，红黄兰，小宝宝，最爱玩，搭火车，钻山洞，盖高楼，连着天。小汽车小汽车，嘀嘀嘀，开过来，开过去，小宝宝，当司机，送妈妈，上班去。藏猫猫儿歌：躲猫猫，躲猫猫， 猫猫、猫猫在哪里？喵……猫咪在这里。",
    ]

    for prompt_audio_path in [
        # "/VITA-Audio/asset/2631296891109983590.wav",
        # "/VITA-Audio/asset/379838640-d5ff0815-74f8-4738-b0f1-477cfc8dcc2d.wav",
        "/VITA-Audio/asset/4202818730519913143.wav",
    ]:
        subprocess.run(f"cp {prompt_audio_path} {ASSETS_DIR}", shell=True)
        file_name = os.path.basename(prompt_audio_path)[:-4]
        for text in TTS_texts:
            print("=" * 100)
            print("clone_tts_task")
            print(f"{text=} {prompt_audio_path=}")

            start = perf_counter()
            output, tts_speech = s2s_inference.run_infer(
                prompt_audio_path=prompt_audio_path,
                # message="Translate the text to speech.\n" + text,
                message="Convert the text to speech.\n" + text,
                mode=None,
                do_sample=True,
            )
            print(f"{output=}", flush=True)
            print(f"{tts_speech.shape=}", flush=True)
            print(f"cost: {perf_counter()-start} s")

            wav_path = os.path.join(ASSETS_DIR, file_name[:16] + "_" + text[:16] + ".wav")
            print(f"save to {wav_path}")
            torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")


# ==============================================================
# TTS stream
def tts_stream():
    """all text gen over, then to gen speech"""
    import torchaudio
    import soundfile as sf

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        flow_path=flow_path,
    )

    TTS_texts = [
        # "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",  # warmup
        # "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",
        # "起床歌：小宝宝，起得早，睁开眼，眯眯笑，咿呀呀，学说话，伸伸手，要人抱。穿衣歌小胳膊，穿袖子，穿上衣，扣扣子，小脚丫，穿裤子，穿上袜子穿鞋子。",
        "小镜子-小镜子，圆又圆，看宝宝，露笑脸。闭上眼，做个梦，变月亮，挂上天。",
        # "小铃铛叮铃铃，叮铃铃，一会远，一会近。小宝宝，耳朵灵，听铃声，找到铃。学画画小宝宝，学画画，大蜡笔，手中拿，画小鸭，叫嘎嘎，画小马，骑回家。大鞋子大鞋子，像只船，爸爸穿，我也穿，一二一，向前走，走呀走，翻了船。",
        # "逛公园逛公园，宝宝笑，东看看，西瞧瞧，花儿香，鸟儿叫，小草绿，小树摇。看画报小娃娃，看画报，睁大眼，仔细瞧，布娃娃，哈哈笑，伸伸手，要你抱。",
        # "搭积木大积木，红黄兰，小宝宝，最爱玩，搭火车，钻山洞，盖高楼，连着天。小汽车小汽车，嘀嘀嘀，开过来，开过去，小宝宝，当司机，送妈妈，上班去。藏猫猫儿歌：躲猫猫，躲猫猫， 猫猫、猫猫在哪里？喵……猫咪在这里。",
    ]

    for text in TTS_texts:
        print("=" * 100)
        print("tts_stream_task")
        print(f"{text=}")

        file_name = text[:16]
        times = []
        start_time = perf_counter()

        generated_text = ""
        audio_decode_time = []
        audios = []
        for new_text in s2s_inference.run_infer_stream(
            message="Convert the text to speech.\n" + text,
            mode=None,
            do_sample=True,
        ):
            times.append(perf_counter() - start_time)
            generated_text += new_text
            print(new_text, end=" ", flush=True)
            start_time = perf_counter()
        print(
            f"\ngenerate first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s\n"
        )

        audio_decode_time = []
        audios = []
        audio_segments = find_audio_segments_regex(generated_text)
        for audio_idx, audio_segment in enumerate(audio_segments):
            start = perf_counter()

            audio_tokens = extract_token_ids_as_int(audio_segment)
            print(f"{audio_tokens=}")

            tts_speech = s2s_inference.audio_tokenizer.decode(audio_tokens)
            audio_decode_time.append(perf_counter() - start)

            # wav_path = os.path.join(ASSETS_DIR, f"qa_stream_{audio_idx}_{file_name}")
            # print(f"save to {wav_path=}")
            # torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")

            audios.append(tts_speech.cpu().numpy())

        wav_path = os.path.join(ASSETS_DIR, f"sts_stream_{file_name}.wav")
        print(f"save to {wav_path=}")
        sf.write(wav_path, np.concatenate(audios), samplerate=22050)
        info = sf.info(wav_path, verbose=True)

        print(
            f"\ngenerate first audio segment cost time: {audio_decode_time[0]} s, {len(audio_decode_time)} segment cost time: {sum(audio_decode_time)} s | wav duration: {info.duration} s | audio tokenizer decode RTF: {sum(audio_decode_time)/info.duration} \n"
        )


# TTS stream
def tts_text_stream():
    import torchaudio
    import soundfile as sf

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        flow_path=flow_path,
    )

    TTS_texts = [
        "你好，hello.",
        # "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",  # warmup
        # "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",
        # "起床歌：小宝宝，起得早，睁开眼，眯眯笑，咿呀呀，学说话，伸伸手，要人抱。穿衣歌小胳膊，穿袖子，穿上衣，扣扣子，小脚丫，穿裤子，穿上袜子穿鞋子。",
        # "小镜子-小镜子，圆又圆，看宝宝，露笑脸。闭上眼，做个梦，变月亮，挂上天。",
        # "小铃铛叮铃铃，叮铃铃，一会远，一会近。小宝宝，耳朵灵，听铃声，找到铃。学画画小宝宝，学画画，大蜡笔，手中拿，画小鸭，叫嘎嘎，画小马，骑回家。大鞋子大鞋子，像只船，爸爸穿，我也穿，一二一，向前走，走呀走，翻了船。",
        # "逛公园逛公园，宝宝笑，东看看，西瞧瞧，花儿香，鸟儿叫，小草绿，小树摇。看画报小娃娃，看画报，睁大眼，仔细瞧，布娃娃，哈哈笑，伸伸手，要你抱。",
        # "搭积木大积木，红黄兰，小宝宝，最爱玩，搭火车，钻山洞，盖高楼，连着天。小汽车小汽车，嘀嘀嘀，开过来，开过去，小宝宝，当司机，送妈妈，上班去。藏猫猫儿歌：躲猫猫，躲猫猫， 猫猫、猫猫在哪里？喵……猫咪在这里。",
    ]

    for text in TTS_texts:
        print("=" * 100)
        print("tts_stream_task")
        print(f"{text=}")

        chunk_size = 4 * 2

        file_name = text[:16]
        times = []
        start_time = perf_counter()

        generated_text = ""
        sub_generated_text = ""
        audio_decode_time = []
        audios = []
        for new_text in s2s_inference.run_infer_stream(
            message="Convert the text to speech.\n" + text,
            mode=None,
            do_sample=True,
        ):
            times.append(perf_counter() - start_time)
            generated_text += new_text
            sub_generated_text += new_text
            print(new_text, end="", flush=True)
            if "<|end_of_audio|>" in new_text:
                audio_segments = find_audio_segments_regex(sub_generated_text)
                for audio_idx, audio_segment in enumerate(audio_segments):
                    audio_tokens = extract_token_ids_as_int(audio_segment)
                    print(f"\n{audio_tokens=}", flush=True)
                    start_time = perf_counter()
                    tts_speech = s2s_inference.audio_tokenizer.decode(audio_tokens)
                    audio_decode_time.append(perf_counter() - start_time)

                    # wav_path = os.path.join(output_dir, file_name + f"_{pos}.wav")
                    # torchaudio.save(wav_path, tts_speech.unsqueeze(0), 22050, format="wav")

                    audios.append(tts_speech.cpu().numpy())
                sub_generated_text = ""
            start_time = perf_counter()
        print(
            f"\ngenerate first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s\n"
        )

        wav_path = os.path.join(ASSETS_DIR, f"sts_stream_{file_name}.wav")
        print(f"save to {wav_path=}")
        sf.write(wav_path, np.concatenate(audios), samplerate=22050)
        info = sf.info(wav_path, verbose=True)

        print(
            f"\ngenerate first audio segment cost time: {audio_decode_time[0]} s, {len(audio_decode_time)} segment cost time: {sum(audio_decode_time)} s | wav duration: {info.duration} s | audio tokenizer decode RTF: {sum(audio_decode_time)/info.duration} \n"
        )


# TTS stream
def tts_audio_chunk_static_stream():
    import torchaudio
    import soundfile as sf

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        flow_path=flow_path,
    )

    TTS_texts = [
        # "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",  # warmup
        # "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",
        # "起床歌：小宝宝，起得早，睁开眼，眯眯笑，咿呀呀，学说话，伸伸手，要人抱。穿衣歌小胳膊，穿袖子，穿上衣，扣扣子，小脚丫，穿裤子，穿上袜子穿鞋子。",
        "小镜子-小镜子，圆又圆，看宝宝，露笑脸。闭上眼，做个梦，变月亮，挂上天。",
        # "小铃铛叮铃铃，叮铃铃，一会远，一会近。小宝宝，耳朵灵，听铃声，找到铃。学画画小宝宝，学画画，大蜡笔，手中拿，画小鸭，叫嘎嘎，画小马，骑回家。大鞋子大鞋子，像只船，爸爸穿，我也穿，一二一，向前走，走呀走，翻了船。",
        # "逛公园逛公园，宝宝笑，东看看，西瞧瞧，花儿香，鸟儿叫，小草绿，小树摇。看画报小娃娃，看画报，睁大眼，仔细瞧，布娃娃，哈哈笑，伸伸手，要你抱。",
        # "搭积木大积木，红黄兰，小宝宝，最爱玩，搭火车，钻山洞，盖高楼，连着天。小汽车小汽车，嘀嘀嘀，开过来，开过去，小宝宝，当司机，送妈妈，上班去。藏猫猫儿歌：躲猫猫，躲猫猫， 猫猫、猫猫在哪里？喵……猫咪在这里。",
    ]

    for i, text in enumerate(TTS_texts):
        print("=" * 100)
        print("tts_stream_task")
        print(f"{text=}")

        chunk_size = 4 * 2
        session_id = i

        file_name = text[:16]
        times = []
        start_time = perf_counter()

        generated_text = ""
        raw_text = ""
        audio_decode_time = []
        audio_chunk = []
        audios = []
        for new_text in s2s_inference.run_infer_stream(
            message="Convert the text to speech.\n" + text,
            mode=None,
            do_sample=True,
        ):
            times.append(perf_counter() - start_time)
            generated_text += new_text
            print(new_text, end="", flush=True)
            if "audio" not in new_text:
                raw_text += new_text
                continue
            audio_tokens = extract_token_ids_as_int(new_text)
            print(f"\n{audio_tokens=}", flush=True)
            if not audio_tokens:
                continue
            audio_chunk.extend(audio_tokens)
            if len(audio_chunk) % chunk_size == 0:
                chunk = audio_chunk
                print(f"\n{chunk=}", flush=True)
                start_time = perf_counter()
                tts_speech = s2s_inference.audio_tokenizer.decode(chunk, session_id=session_id)
                audio_decode_time.append(perf_counter() - start_time)

                audios.append(tts_speech.cpu().numpy())
                audio_chunk = []
                start_time = perf_counter()

        if len(audio_chunk) > 0:
            chunk = audio_chunk
            print(f"\nlast {chunk=}", flush=True)
            start_time = perf_counter()
            tts_speech = s2s_inference.audio_tokenizer.decode(chunk, session_id=session_id)
            audio_decode_time.append(perf_counter() - start_time)

            audios.append(tts_speech.cpu().numpy())

        print(
            f"\ngenerate [{raw_text}] first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s\n"
        )

        wav_path = os.path.join(ASSETS_DIR, f"sts_stream_{file_name}.wav")
        print(f"save to {wav_path=}")
        sf.write(wav_path, np.concatenate(audios), samplerate=22050)
        info = sf.info(wav_path, verbose=True)

        print(
            f"\ngenerate first audio segment cost time: {audio_decode_time[0]} s, {len(audio_decode_time)} segment cost time: {sum(audio_decode_time)} s | wav duration: {info.duration} s | audio tokenizer decode RTF: {sum(audio_decode_time)/info.duration} \n"
        )


def tts_audio_chunk_dynamic_stream():
    """
    smooth chunk tokens_decode_out from glm4voice: chunk_size_list = [25, 50, 100, 150, 200]
    """
    import torchaudio
    import soundfile as sf

    s2s_inference = S2SInference(
        model_path,
        audio_tokenizer_type,
        flow_path=flow_path,
    )

    TTS_texts = [
        # "你好，hello.",
        "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",  # warmup
        # "他本是我莲花池里养大的金鱼，每日浮头听经，修成手段。那一柄九瓣铜锤，乃是一枝未开的菡萏，被他运炼成兵。不知是那一日，海潮泛涨，走到此间。我今早扶栏看花，却不见这厮出拜，掐指巡纹，算着他在此成精，害你师父，故此未及梳妆，运神功，织个竹篮儿擒他。",
        # "起床歌：小宝宝，起得早，睁开眼，眯眯笑，咿呀呀，学说话，伸伸手，要人抱。穿衣歌小胳膊，穿袖子，穿上衣，扣扣子，小脚丫，穿裤子，穿上袜子穿鞋子。",
        # "小镜子-小镜子，圆又圆，看宝宝，露笑脸。闭上眼，做个梦，变月亮，挂上天。",
        # "小铃铛叮铃铃，叮铃铃，一会远，一会近。小宝宝，耳朵灵，听铃声，找到铃。学画画小宝宝，学画画，大蜡笔，手中拿，画小鸭，叫嘎嘎，画小马，骑回家。大鞋子大鞋子，像只船，爸爸穿，我也穿，一二一，向前走，走呀走，翻了船。",
        # "逛公园逛公园，宝宝笑，东看看，西瞧瞧，花儿香，鸟儿叫，小草绿，小树摇。看画报小娃娃，看画报，睁大眼，仔细瞧，布娃娃，哈哈笑，伸伸手，要你抱。",
        # "搭积木大积木，红黄兰，小宝宝，最爱玩，搭火车，钻山洞，盖高楼，连着天。小汽车小汽车，嘀嘀嘀，开过来，开过去，小宝宝，当司机，送妈妈，上班去。藏猫猫儿歌：躲猫猫，躲猫猫， 猫猫、猫猫在哪里？喵……猫咪在这里。",
    ]

    for i, text in enumerate(TTS_texts):
        print("=" * 100)
        print("tts_stream_task")
        print(f"{text=}")

        prompt_speech_feat = torch.zeros(1, 0, 80).to("cuda")
        flow_prompt_speech_token = torch.zeros(1, 0, dtype=torch.int64).to("cuda")

        this_uuid = str(uuid.uuid4())
        tts_mels = []
        prev_mel = None

        is_finalize = False
        chunk_size_list = [8, 16, 25, 50, 100, 150, 200]  # like tcp cwnd
        # chunk_size_list = [8]  # like tcp cwnd
        chunk_size_idx = 0
        chunk_size = chunk_size_list[chunk_size_idx]

        file_name = text[:16]
        times = []
        start_time = perf_counter()

        generated_text = ""
        raw_text = ""
        audio_decode_time = []
        audio_chunk = []
        audios = []
        for new_text in s2s_inference.run_infer_stream(
            message="Convert the text to speech.\n" + text,
            mode=None,
            do_sample=True,
        ):
            times.append(perf_counter() - start_time)
            generated_text += new_text
            print(f"{new_text=}", flush=True)
            if "<|begin_of_audio|>" in new_text:
                new_text = new_text.replace("<|begin_of_audio|>", "")
            if "<|end_of_audio|>" in new_text:
                new_text = new_text.replace("<|end_of_audio|>", "")
            if "<|im_end|>" in new_text:
                new_text = new_text.replace("<|im_end|>", "")
                is_finalize = True
            audio_tokens = extract_token_ids_as_int(new_text)
            print(f"\n{audio_tokens=}", flush=True)
            if not audio_tokens and is_finalize is False:
                raw_text += new_text
                continue
            audio_chunk.extend(audio_tokens)
            print(f"{is_finalize=} {len(audio_chunk)=}")
            if len(audio_chunk) >= chunk_size or (is_finalize and audio_chunk):
                print(f"\n{audio_chunk=}", flush=True)
                if chunk_size_idx < len(chunk_size_list) - 1:
                    chunk_size_idx += 1
                    chunk_size = chunk_size_list[chunk_size_idx]
                tts_token = torch.tensor(audio_chunk, device="cuda").unsqueeze(0)

                if prev_mel is not None:
                    prompt_speech_feat = torch.cat(tts_mels, dim=-1).transpose(1, 2)

                # gen waveform and mel-spectrogram feat
                start_time = perf_counter()
                tts_speech, tts_mel = s2s_inference.audio_tokenizer.audio_decoder.token2wav(
                    tts_token,
                    uuid=this_uuid,
                    prompt_token=flow_prompt_speech_token.to("cuda"),
                    prompt_feat=prompt_speech_feat.to("cuda"),
                    finalize=is_finalize,
                )
                audio_decode_time.append(perf_counter() - start_time)
                prev_mel = tts_mel

                print(tts_speech.shape)
                audios.append(tts_speech.squeeze(0).cpu().numpy())  # T
                tts_mels.append(tts_mel)
                flow_prompt_speech_token = torch.cat((flow_prompt_speech_token, tts_token), dim=-1)
                audio_chunk = []
                start_time = perf_counter()

        print(
            f"\ngenerate [{raw_text}] first token cost time: {times[0]} s, {len(times)} tokens cost time: {sum(times)} s\n"
        )
        print(f"{chunk_size_list=}")
        print(f"{audio_decode_time=}")

        wav_path = os.path.join(ASSETS_DIR, f"sts_stream_{file_name}.wav")
        print(f"save to {wav_path=}")
        sf.write(wav_path, np.concatenate(audios), samplerate=22050)
        info = sf.info(wav_path, verbose=True)

        print(
            f"\ngenerate first audio segment cost time: {audio_decode_time[0]} s, {len(audio_decode_time)} segment cost time: {sum(audio_decode_time)} s | wav duration: {info.duration} s | audio tokenizer decode RTF: {sum(audio_decode_time)/info.duration} \n"
        )


def print_model_params(model: torch.nn.Module, extra_info=""):
    # print the number of parameters in the model
    model_million_params = sum(p.numel() for p in model.parameters()) / 1e6
    print(model)
    print(f"{extra_info} {model_million_params} M parameters")


def find_audio_segments_regex(text):
    """
    Find all substrings between <|begin_of_audio|> and <|end_of_audio|> using regex.

    Args:
        text (str): The input string to search through

    Returns:
        list: A list of all found audio segments (substrings between the delimiters)
    """
    pattern = re.compile(r"<\|begin_of_audio\|>(.*?)<\|end_of_audio\|>", re.DOTALL)
    segments = pattern.findall(text)
    return [segment.strip() for segment in segments]


def extract_token_ids_as_int(text):
    pattern = re.compile(r"<\|audio_(\d+)\|>")
    token_ids = pattern.findall(text)
    return [int(id) for id in token_ids]


def custom_init_weights(module):
    if isinstance(module, torch.nn.Conv2d) or isinstance(module, torch.nn.Linear):
        torch.nn.init.xavier_uniform_(module.weight)
        if module.bias is not None:
            torch.nn.init.constant_(module.bias, 0)
    elif isinstance(module, torch.nn.BatchNorm2d) or isinstance(module, torch.nn.BatchNorm1d):
        torch.nn.init.constant_(module.weight, 1)
        torch.nn.init.constant_(module.bias, 0)


class TextAudioIteratorStreamer(TextIteratorStreamer):
    def __init__(
        self,
        tokenizer: "AutoTokenizer",
        skip_prompt: bool = False,
        timeout: Optional[float] = None,
        **decode_kwargs,
    ):
        super().__init__(tokenizer, skip_prompt, timeout, **decode_kwargs)

        # self.audio_offset = tokenizer.convert_tokens_to_ids("<|audio_0|>")
        self.audio_offset = tokenizer.convert_tokens_to_ids("<|begin_of_audio|>")
        self.num_decode_tokens = 0

    def put(self, value):
        """
        Receives tokens, decodes them, and prints them to stdout as soon as they form entire words.
        """
        print(f"{value=}")
        if len(value.shape) > 1 and value.shape[0] > 1:
            raise ValueError("TextStreamer only supports batch size 1")
        elif len(value.shape) > 1:
            value = value[0]

        if self.skip_prompt and self.next_tokens_are_prompt:
            self.next_tokens_are_prompt = False
            return

        self.num_decode_tokens += len(value)

        # Add the new token to the cache and decodes the entire thing.
        self.token_cache.extend(value.tolist())
        text = self.tokenizer.decode(self.token_cache, **self.decode_kwargs)

        # After the symbol for a new line, we flush the cache.
        if text.endswith("\n"):
            printable_text = text[self.print_len :]
            self.token_cache = []
            self.print_len = 0
        # If the last token is a CJK character, we print the characters.
        elif len(text) > 0 and self._is_chinese_char(ord(text[-1])):
            printable_text = text[self.print_len :]
            self.print_len += len(printable_text)
        elif self.token_cache[-1] >= self.audio_offset:
            printable_text = text[self.print_len :]
            self.print_len += len(printable_text)
        # Otherwise, prints until the last space char (simple heuristic to avoid printing incomplete words,
        # which may change with the subsequent token -- there are probably smarter ways to do this!)
        else:
            printable_text = text[self.print_len : text.rfind(" ") + 1]
            self.print_len += len(printable_text)

        self.on_finalized_text(printable_text)

        while self.text_queue.qsize() > 10:
            time.sleep(0.01)


class BenchmarkIteratorStreamer(TextIteratorStreamer):
    def __init__(
        self,
        tokenizer: "AutoTokenizer",
        skip_prompt: bool = False,
        timeout: Optional[float] = None,
        **decode_kwargs,
    ):
        super().__init__(tokenizer, skip_prompt, timeout, **decode_kwargs)

        self.num_decode_tokens = 0

    def put(self, value):
        """
        Receives tokens, decodes them, and prints them to stdout as soon as they form entire words.
        """
        if len(value.shape) > 1 and value.shape[0] > 1:
            raise ValueError("TextStreamer only supports batch size 1")
        elif len(value.shape) > 1:
            value = value[0]

        if self.skip_prompt and self.next_tokens_are_prompt:
            self.next_tokens_are_prompt = False
            return

        self.num_decode_tokens += len(value)

        printable_text = " ".join([str(x) for x in value.tolist()]) + " "
        self.on_finalized_text(printable_text)


class S2SInference:
    def __init__(
        self,
        llm_model_path,
        audio_tokenizer_type,
        audio_tokenizer_model_path=None,
        sense_voice_model_path=None,
        flow_path=None,
        device_map="cuda:0",
        torch_dtype=torch.bfloat16,
        audio_tokenizer_rank=0,
    ):
        from evaluation.get_chat_template import qwen2_chat_template as chat_template

        add_generation_prompt = True

        default_system_message = []

        luke_system_message = [
            {
                "role": "system",
                "content": "Your Name: Luke\nYour Gender: male\n\nRespond in a text-audio interleaved manner.",
            },
        ]

        tokenizer = AutoTokenizer.from_pretrained(
            llm_model_path,
            trust_remote_code=True,
            chat_template=chat_template,
        )

        # https://huggingface.co/VITA-MLLM/VITA-Audio-Boost/blob/main/modeling_qwen2.py#L781
        # https://huggingface.co/VITA-MLLM/VITA-Audio-Balance/blob/main/modeling_qwen2.py#L781
        # https://huggingface.co/VITA-MLLM/VITA-Audio-Plus-Vanilla/blob/main/modeling_qwen2.py#L834
        model = AutoModelForCausalLM.from_pretrained(
            llm_model_path,
            trust_remote_code=True,
            device_map=device_map,
            torch_dtype=torch_dtype,
            attn_implementation="flash_attention_2",
        ).eval()
        # print("model", model)
        print(f"{model.config=}")
        # print(f"{model.config.model_type=}")
        print(f"{model.hf_device_map=}")

        model.generation_config = GenerationConfig.from_pretrained(
            llm_model_path, trust_remote_code=True
        )

        model.generation_config.max_new_tokens = 8192
        model.generation_config.chat_format = "chatml"
        model.generation_config.max_window_size = 8192
        model.generation_config.use_cache = True
        # model.generation_config.use_cache = False
        model.generation_config.do_sample = False
        model.generation_config.temperature = 1.0
        model.generation_config.top_k = 50
        model.generation_config.top_p = 1.0
        model.generation_config.num_beams = 1
        model.generation_config.pad_token_id = tokenizer.pad_token_id
        print(f"{model.generation_config=}")

        audio_tokenizer = get_audio_tokenizer(
            model_type=audio_tokenizer_type,
            flow_path=flow_path,
            model_name_or_path=audio_tokenizer_model_path,
            rank=audio_tokenizer_rank,
            sense_voice_model_path=sense_voice_model_path,
        )
        audio_tokenizer.load_model()

        self.model = model
        self.tokenizer = tokenizer
        self.audio_tokenizer = audio_tokenizer
        self.add_generation_prompt = add_generation_prompt
        self.default_system_message = default_system_message
        self.luke_system_message = luke_system_message

        self.audio_0_id = tokenizer("<|audio_0|>").input_ids[0]
        print(f"{self.audio_0_id=}")

    def benchmark_forward(self, mtp_inference_mode):
        print("-" * 100)
        print("benchmark_forward...")
        print(f"{mtp_inference_mode=}")

        total_time = 0

        past_key_values = None
        use_cache = True

        self.model.input_ids = None
        self.model.inputs_embeds = None
        self.model.hidden_states = [None] * (self.model.config.num_nextn_predict_layers + 1)
        self.model.position_ids = None
        self.model.attention_mask = None
        self.model.mtp_idx = -1
        self.model.num_prefill_tokens = -1

        model_max_length = 1024
        if mtp_inference_mode is not None:
            ori_mtp_inference_mode = self.model.generation_config.mtp_inference_mode
            self.model._prepare_mtp_for_generation(mtp_inference_mode, model_max_length)

        else:
            self.model._prepare_mtp_for_generation(
                self.model.generation_config.mtp_inference_mode, model_max_length
            )

        for i in tqdm.tqdm(range(1, model_max_length + 1)):
            if use_cache:
                input_ids = torch.tensor([i - 1], dtype=torch.long).unsqueeze(0).to("cuda")
                position_ids = torch.tensor([i - 1], dtype=torch.long).unsqueeze(0).to("cuda")
            else:
                input_ids = torch.arange(i, dtype=torch.long).unsqueeze(0).to("cuda")
                position_ids = torch.arange(i, dtype=torch.long).unsqueeze(0).to("cuda")

            attention_mask = torch.tensor([1] * i, dtype=torch.float).unsqueeze(0).to("cuda")

            torch.cuda.synchronize()
            start = time.time()

            output = self.model(
                input_ids,
                attention_mask=attention_mask,
                position_ids=position_ids,
                past_key_values=past_key_values,
                use_cache=use_cache,
                num_logits_to_keep=1,
            )

            torch.cuda.synchronize()
            end = time.time()

            total_time += end - start
            # print(f"{i=} {total_time=}")

            past_key_values = output.past_key_values

        print()
        print(f"{total_time=}")
        print(f"second/token {total_time/model_max_length=}")
        print(f"token/second {model_max_length/total_time=}")

        if mtp_inference_mode is not None:
            self.model.mtp_inference_mode = ori_mtp_inference_mode

    def benchmark_generate(self, mtp_inference_mode):
        self.model.apply(custom_init_weights)

        print("-" * 100)
        print("benchmark_generate...")
        print(f"{mtp_inference_mode=}")

        total_time = 0
        self.model.generation_config.use_cache = True

        self.model.generation_config.max_new_tokens = 8192

        if mtp_inference_mode is not None:
            ori_mtp_inference_mode = self.model.generation_config.mtp_inference_mode
            self.model.generation_config.mtp_inference_mode = mtp_inference_mode

        input_ids = torch.tensor([0], dtype=torch.long).unsqueeze(0).to("cuda")

        torch.cuda.synchronize()
        start = time.time()

        output = self.model.generate(
            input_ids,
        )
        # print(f"{output.size()=}")

        torch.cuda.synchronize()
        end = time.time()

        total_time += end - start

        print()
        print(f"{total_time=}")
        print(f"second/token {total_time/output.size(1)=}")
        print(f"token/second {output.size(1)/total_time=}")

        if mtp_inference_mode is not None:
            self.model.generation_config.mtp_inference_mode = ori_mtp_inference_mode

    def benchmark_generate_stream(self, mtp_inference_mode):
        print("-" * 100)
        print("benchmark_generate_stream...")
        print(f"{mtp_inference_mode=}")

        self.model.apply(custom_init_weights)

        total_time = 0
        self.model.generation_config.use_cache = True

        # model_max_length = 8192
        model_max_length = 4096
        # model_max_length = 2048
        # model_max_length = 1024
        num_prefill_tokens = 32

        self.model.generation_config.max_new_tokens = model_max_length
        self.model.generation_config.do_sample = False

        if mtp_inference_mode is not None:
            ori_mtp_inference_mode = self.model.generation_config.mtp_inference_mode
            self.model.generation_config.mtp_inference_mode = mtp_inference_mode

        input_ids = torch.tensor([0] * num_prefill_tokens, dtype=torch.long).unsqueeze(0).to("cuda")

        streamer = BenchmarkIteratorStreamer(self.tokenizer, skip_prompt=True)
        generation_kwargs = dict(input_ids=input_ids, streamer=streamer)
        thread = Thread(target=self.model.generate, kwargs=generation_kwargs)

        token_decode_time = []

        torch.cuda.synchronize()
        start = time.time()
        thread.start()

        generated_text = ""
        for new_text in tqdm.tqdm(streamer, total=model_max_length):
            generated_text += new_text
            end = time.time()

            token_decode_time.append(end - start)

            yield new_text

        # print(f"{len(generated_text)}")

        torch.cuda.synchronize()
        end = time.time()

        total_time += end - start

        print()
        print(f"{token_decode_time[-1]=}")
        print(f"{streamer.num_decode_tokens=}")
        print(f"second/token {token_decode_time[-1]/streamer.num_decode_tokens=}")
        print(f"token/second {streamer.num_decode_tokens/token_decode_time[-1]=}")

        # if mtp_inference_mode is None:
        #     mtp_inference_mode = []
        # with open(f'token_decode_time_{str(mtp_inference_mode)}.json', 'w') as f:
        #     json.dump(token_decode_time, f)

        if mtp_inference_mode is not None:
            self.model.generation_config.mtp_inference_mode = ori_mtp_inference_mode

    @torch.inference_mode()
    def run_infer(
        self,
        audio_path=None,
        prompt_audio_path=None,
        chunk_size=4,
        max_returned_tokens=4096,
        sample_rate=16000,
        request_id="",
        audio_feats=None,
        message="",
        use_past=False,
        mode="luke",
        do_sample=False,
        mtp_inference_mode=None,
    ):
        AUD_TAG_TOKEN = "<|audio|>"
        AUD_CONTEXT_TOKEN = "<|context_of_audio|>"
        AUD_START_TOKEN = "<|begin_of_audio|>"
        AUD_END_TOKEN = "<|end_of_audio|>"

        if prompt_audio_path is not None:
            system_message = [
                {
                    "role": "system",
                    "content": f"Your Voice: <|audio|>\n",
                },
            ]

        elif mode == "luke":
            system_message = self.luke_system_message

        else:
            system_message = self.default_system_message

        if prompt_audio_path is not None and self.audio_tokenizer.apply_to_role(
            "user", is_discrete=True
        ):
            print("discrete codec")
            # discrete codec
            audio_tokens = self.audio_tokenizer.encode(
                prompt_audio_path, is_discrete=True, is_contiguous=False
            )
            audio_tokens = "".join(f"<|audio_{i}|>" for i in audio_tokens)
            system_message[-1]["content"] = system_message[-1]["content"].replace(
                "<|audio|>", f"<|begin_of_audio|>{audio_tokens}<|end_of_audio|>"
            )

        if audio_path is not None:
            messages = system_message + [
                {
                    "role": "user",
                    "content": message + "\n<|audio|>",
                },
            ]
        else:
            messages = system_message + [
                {
                    "role": "user",
                    "content": message,
                },
            ]

        if audio_path is not None and self.audio_tokenizer.apply_to_role("user", is_discrete=True):
            print("discrete codec")
            # discrete codec
            audio_tokens = self.audio_tokenizer.encode(
                audio_path, is_discrete=True, is_contiguous=False
            )
            audio_tokens = "".join(f"<|audio_{i}|>" for i in audio_tokens)
            messages[-1]["content"] = messages[-1]["content"].replace(
                "<|audio|>", f"<|begin_of_audio|>{audio_tokens}<|end_of_audio|>"
            )

        input_ids = self.tokenizer.apply_chat_template(
            messages,
            tokenize=True,
            add_generation_prompt=self.add_generation_prompt,
        )

        if (
            audio_path is not None or prompt_audio_path is not None
        ) and self.audio_tokenizer.apply_to_role("user", is_contiguous=True):
            print("contiguous codec")
            # contiguous codec
            audio_paths = []
            if audio_path is not None:
                audio_paths.append(audio_path)
            if prompt_audio_path is not None:
                audio_paths.append(prompt_audio_path)
            input_ids, audios, audio_indices = add_audio_input_contiguous(
                input_ids, audio_paths, self.tokenizer, self.audio_tokenizer
            )
        else:
            audios = None
            audio_indices = None

        input_ids = torch.tensor([input_ids], dtype=torch.long).to("cuda")
        print("input", self.tokenizer.decode(input_ids[0], skip_special_tokens=False), flush=True)

        self.model.generation_config.do_sample = do_sample

        if mtp_inference_mode is not None:
            ori_mtp_inference_mode = self.model.generation_config.mtp_inference_mode
            self.model.generation_config.mtp_inference_mode = mtp_inference_mode

        outputs = self.model.generate(
            input_ids,
            audios=audios,
            audio_indices=audio_indices,
        )

        outputs = outputs[0][input_ids.shape[1] :]
        output = self.tokenizer.decode(outputs, skip_special_tokens=False)

        audio_offset = self.tokenizer.convert_tokens_to_ids("<|audio_0|>")

        audio_tokens = []
        for token_id in outputs:
            if token_id >= audio_offset:
                audio_tokens.append(token_id - audio_offset)

        if len(audio_tokens) > 0:
            tts_speech = self.audio_tokenizer.decode(
                audio_tokens, source_speech_16k=prompt_audio_path
            )

        else:
            tts_speech = None

        if mtp_inference_mode is not None:
            self.model.generation_config.mtp_inference_mode = ori_mtp_inference_mode

        return output, tts_speech

    @torch.inference_mode()
    def run_infer_stream(
        self,
        audio_path=None,
        prompt_audio_path=None,
        chunk_size=4,
        max_returned_tokens=4096,
        sample_rate=16000,
        request_id="",
        audio_feats=None,
        message="",
        use_past=False,
        mode="luke",
        do_sample=False,
        mtp_inference_mode=None,
    ):
        if prompt_audio_path is not None:
            system_message = [
                {
                    "role": "system",
                    "content": f"Your Voice: <|audio|>\n",
                },
            ]

        elif mode == "luke":
            system_message = self.luke_system_message

        else:
            system_message = self.default_system_message

        if prompt_audio_path is not None and self.audio_tokenizer.apply_to_role(
            "user", is_discrete=True
        ):
            # discrete codec
            audio_tokens = self.audio_tokenizer.encode(prompt_audio_path)
            audio_tokens = "".join(f"<|audio_{i}|>" for i in audio_tokens)
            system_message[-1]["content"] = system_message[-1]["content"].replace(
                "<|audio|>", f"<|begin_of_audio|>{audio_tokens}<|end_of_audio|>"
            )

        if audio_path is not None:
            messages = system_message + [
                {
                    "role": "user",
                    "content": message + "\n<|audio|>",
                },
            ]
        else:
            messages = system_message + [
                {
                    "role": "user",
                    "content": message,
                },
            ]

        if audio_path is not None and self.audio_tokenizer.apply_to_role("user", is_discrete=True):
            print("discrete codec")
            # discrete codec
            audio_tokens = self.audio_tokenizer.encode(audio_path)
            audio_tokens = "".join(f"<|audio_{i}|>" for i in audio_tokens)
            messages[-1]["content"] = messages[-1]["content"].replace(
                "<|audio|>", f"<|begin_of_audio|>{audio_tokens}<|end_of_audio|>"
            )

        input_ids = self.tokenizer.apply_chat_template(
            messages,
            tokenize=True,
            add_generation_prompt=self.add_generation_prompt,
        )

        if (
            audio_path is not None or prompt_audio_path is not None
        ) and self.audio_tokenizer.apply_to_role("user", is_contiguous=True):
            print("contiguous codec")
            # contiguous codec
            audio_paths = []
            if audio_path is not None:
                audio_paths.append(audio_path)
            if prompt_audio_path is not None:
                audio_paths.append(prompt_audio_path)
            input_ids, audios, audio_indices = add_audio_input_contiguous(
                input_ids, audio_paths, self.tokenizer, self.audio_tokenizer
            )
        else:
            audios = None
            audio_indices = None

        input_ids = torch.tensor([input_ids], dtype=torch.long).to("cuda")

        print("input", self.tokenizer.decode(input_ids[0], skip_special_tokens=False), flush=True)
        attention_mask = torch.ones((1, input_ids.shape[1]), dtype=torch.int64).to("cuda")

        self.model.generation_config.do_sample = do_sample

        if mtp_inference_mode is not None:
            print(f"{mtp_inference_mode=}")
            ori_mtp_inference_mode = self.model.generation_config.mtp_inference_mode
            self.model.generation_config.mtp_inference_mode = mtp_inference_mode

        streamer = TextAudioIteratorStreamer(self.tokenizer, skip_prompt=True)
        generation_kwargs = dict(
            input_ids=input_ids,
            attention_mask=attention_mask,
            audios=audios,
            audio_indices=audio_indices,
            streamer=streamer,
        )
        print(
            f"input_ids: {input_ids.shape}, attention_mask: {attention_mask.shape} audio_indices: {audio_indices} audios: {audios}"
        )
        thread = Thread(target=self.model.generate, kwargs=generation_kwargs)

        thread.start()

        # generated_text = ""
        for new_text in streamer:
            # generated_text += new_text

            yield new_text

        # torch.cuda.synchronize()

        if mtp_inference_mode is not None:
            self.model.generation_config.mtp_inference_mode = ori_mtp_inference_mode


"""
IMAGE_GPU=T4 modal run src/llm/transformers/vita_voice.py --task tokenize
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task dump_model

# LLM: sensevoice(no use ctc)_qwen2_mtp(no mtp)
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text 
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text_audio
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text_stream
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text_audio_stream

# LLM: qwen2_mtp
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text 
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text_audio
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text_stream
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text_audio_stream
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text 
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task text_stream

# LLM(Plus-Vanilla): sensevoice(no use ctc)_qwen2_mtp(no mtp) + AudioTokenizer: sensevoice_glm4voice(sensevoice WavFrontend, decoder)
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task sts
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task sts_stream
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task asr
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task asr_text_stream
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_stream
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_clone
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_audio_chunk_static_stream
IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_audio_chunk_dynamic_stream

# LLM(Plus-Boost): sensevoice(no use ctc)_qwen2_mtp(no mtp) + AudioTokenizer: sensevoice_glm4voice(sensevoice WavFrontend, decoder)
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Plus-Boost IMAGE_GPU=L40s modal run src/llm/transformers/vita_voice.py --task sts
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Plus-Boost IMAGE_GPU=L40s modal run src/llm/transformers/vita_voice.py --task sts_stream
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Plus-Boost IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task asr
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Plus-Boost IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task asr_text_stream
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Plus-Boost IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Plus-Boost IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_stream
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Plus-Boost IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_clone
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Plus-Boost IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_audio_chunk_static_stream
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Plus-Boost IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_audio_chunk_dynamic_stream

# LLM(Boost/Balance): qwen2_mtp + AudioTokenizer: glm4voice(tokenizer, decoder)
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L40s modal run src/llm/transformers/vita_voice.py --task sts
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L40s modal run src/llm/transformers/vita_voice.py --task sts_stream
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task asr
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task asr_text_stream
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_stream
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_clone
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_audio_chunk_static_stream
AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_audio_chunk_dynamic_stream

MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L40s modal run src/llm/transformers/vita_voice.py --task sts
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L40s modal run src/llm/transformers/vita_voice.py --task sts_stream
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task asr
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task asr_text_stream
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_stream
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_clone
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_audio_chunk_static_stream
MTP_LLM_MODEL=VITA-MLLM/VITA-Audio-Boost AUDIO_TOKENIZER_TYPE=glm4voice IMAGE_GPU=L4 modal run src/llm/transformers/vita_voice.py --task tts_audio_chunk_dynamic_stream
"""


@app.local_entrypoint()
def main(task: str = "tokenize"):
    tasks = {
        "tokenize": tokenize,
        "dump_model": dump_model,
        "text": text,
        "text_stream": text_stream,
        "text_audio": text_audio,
        "text_audio_stream": text_audio_stream,
        "asr": asr,
        "asr_text_stream": asr_text_stream,
        "sts": sts,
        "sts_stream": sts_stream,
        "tts": tts,
        "tts_stream": tts_stream,
        "tts_clone": tts_clone,
        "tts_audio_chunk_static_stream": tts_audio_chunk_static_stream,
        "tts_audio_chunk_dynamic_stream": tts_audio_chunk_dynamic_stream,
        "benchmark_llm": benchmark_llm,
        # "benchmark_sts": benchmark_sts,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    run.remote(tasks[task])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re, sys, unicodedata
import codecs

remove_tag = True
spacelist = [' ', '\t', '\r', '\n']
puncts = [
    '!', ',', '?', '、', '。', '！', '，', '；', '？', '：', '「', '」', '︰', '『', '』',
    '《', '》'
]


def characterize(string):
    res = []
    i = 0
    while i < len(string):
        char = string[i]
        if char in puncts:
            i += 1
            continue
        cat1 = unicodedata.category(char)
        #https://unicodebook.readthedocs.io/unicode.html#unicode-categories
        if cat1 == 'Zs' or cat1 == 'Cn' or char in spacelist:  # space or not assigned
            i += 1
            continue
        if cat1 == 'Lo':  # letter-other
            res.append(char)
            i += 1
        else:
            # some input looks like: <unk><noise>, we want to separate it to two words.
            sep = ' '
            if char == '<': sep = '>'
            j = i + 1
            while j < len(string):
                c = string[j]
                if ord(c) >= 128 or (c in spacelist) or (c == sep):
                    break
                j += 1
            if j < len(string) and string[j] == '>':
                j += 1
            res.append(string[i:j])
            i = j
    return res


def stripoff_tags(x):
    if not x: return ''
    chars = []
    i = 0
    T = len(x)
    while i < T:
        if x[i] == '<':
            while i < T and x[i] != '>':
                i += 1
            i += 1
        else:
            chars.append(x[i])
            i += 1
    return ''.join(chars)


def normalize(sentence, ignore_words, cs, split=None):
    """ sentence, ignore_words are both in unicode
    """
    new_sentence = []
    for token in sentence:
        x = token
        if not cs:
            x = x.upper()
        if x in ignore_words:
            continue
        if remove_tag:
            x = stripoff_tags(x)
        if not x:
            continue
        if split and x in split:
            new_sentence += split[x]
        else:
            new_sentence.append(x)
    return new_sentence


class Calculator:

    def __init__(self):
        self.data = {}
        self.space = []
        self.cost = {}
        self.cost['cor'] = 0
        self.cost['sub'] = 1
        self.cost['del'] = 1
        self.cost['ins'] = 1

    def calculate(self, lab, rec):
        # Initialization
        lab.insert(0, '')
        rec.insert(0, '')
        while len(self.space) < len(lab):
            self.space.append([])
        for row in self.space:
            for element in row:
                element['dist'] = 0
                element['error'] = 'non'
            while len(row) < len(rec):
                row.append({'dist': 0, 'error': 'non'})
        for i in range(len(lab)):
            self.space[i][0]['dist'] = i
            self.space[i][0]['error'] = 'del'
        for j in range(len(rec)):
            self.space[0][j]['dist'] = j
            self.space[0][j]['error'] = 'ins'
        self.space[0][0]['error'] = 'non'
        for token in lab:
            if token not in self.data and len(token) > 0:
                self.data[token] = {
                    'all': 0,
                    'cor': 0,
                    'sub': 0,
                    'ins': 0,
                    'del': 0
                }
        for token in rec:
            if token not in self.data and len(token) > 0:
                self.data[token] = {
                    'all': 0,
                    'cor': 0,
                    'sub': 0,
                    'ins': 0,
                    'del': 0
                }
        # Computing edit distance
        for i, lab_token in enumerate(lab):
            for j, rec_token in enumerate(rec):
                if i == 0 or j == 0:
                    continue
                min_dist = sys.maxsize
                min_error = 'none'
                dist = self.space[i - 1][j]['dist'] + self.cost['del']
                error = 'del'
                if dist < min_dist:
                    min_dist = dist
                    min_error = error
                dist = self.space[i][j - 1]['dist'] + self.cost['ins']
                error = 'ins'
                if dist < min_dist:
                    min_dist = dist
                    min_error = error
                if lab_token == rec_token:
                    dist = self.space[i - 1][j - 1]['dist'] + self.cost['cor']
                    error = 'cor'
                else:
                    dist = self.space[i - 1][j - 1]['dist'] + self.cost['sub']
                    error = 'sub'
                if dist < min_dist:
                    min_dist = dist
                    min_error = error
                self.space[i][j]['dist'] = min_dist
                self.space[i][j]['error'] = min_error
        # Tracing back
        result = {
            'lab': [],
            'rec': [],
            'all': 0,
            'cor': 0,
            'sub': 0,
            'ins': 0,
            'del': 0
        }
        i = len(lab) - 1
        j = len(rec) - 1
        while True:
            if self.space[i][j]['error'] == 'cor':  # correct
                if len(lab[i]) > 0:
                    self.data[lab[i]]['all'] = self.data[lab[i]]['all'] + 1
                    self.data[lab[i]]['cor'] = self.data[lab[i]]['cor'] + 1
                    result['all'] = result['all'] + 1
                    result['cor'] = result['cor'] + 1
                result['lab'].insert(0, lab[i])
                result['rec'].insert(0, rec[j])
                i = i - 1
                j = j - 1
            elif self.space[i][j]['error'] == 'sub':  # substitution
                if len(lab[i]) > 0:
                    self.data[lab[i]]['all'] = self.data[lab[i]]['all'] + 1
                    self.data[lab[i]]['sub'] = self.data[lab[i]]['sub'] + 1
                    result['all'] = result['all'] + 1
                    result['sub'] = result['sub'] + 1
                result['lab'].insert(0, lab[i])
                result['rec'].insert(0, rec[j])
                i = i - 1
                j = j - 1
            elif self.space[i][j]['error'] == 'del':  # deletion
                if len(lab[i]) > 0:
                    self.data[lab[i]]['all'] = self.data[lab[i]]['all'] + 1
                    self.data[lab[i]]['del'] = self.data[lab[i]]['del'] + 1
                    result['all'] = result['all'] + 1
                    result['del'] = result['del'] + 1
                result['lab'].insert(0, lab[i])
                result['rec'].insert(0, "")
                i = i - 1
            elif self.space[i][j]['error'] == 'ins':  # insertion
                if len(rec[j]) > 0:
                    self.data[rec[j]]['ins'] = self.data[rec[j]]['ins'] + 1
                    result['ins'] = result['ins'] + 1
                result['lab'].insert(0, "")
                result['rec'].insert(0, rec[j])
                j = j - 1
            elif self.space[i][j]['error'] == 'non':  # starting point
                break
            else:  # shouldn't reach here
                print(
                    'this should not happen , i = {i} , j = {j} , error = {error}'
                    .format(i=i, j=j, error=self.space[i][j]['error']))
        return result

    def overall(self):
        result = {'all': 0, 'cor': 0, 'sub': 0, 'ins': 0, 'del': 0}
        for token in self.data:
            result['all'] = result['all'] + self.data[token]['all']
            result['cor'] = result['cor'] + self.data[token]['cor']
            result['sub'] = result['sub'] + self.data[token]['sub']
            result['ins'] = result['ins'] + self.data[token]['ins']
            result['del'] = result['del'] + self.data[token]['del']
        return result

    def cluster(self, data):
        result = {'all': 0, 'cor': 0, 'sub': 0, 'ins': 0, 'del': 0}
        for token in data:
            if token in self.data:
                result['all'] = result['all'] + self.data[token]['all']
                result['cor'] = result['cor'] + self.data[token]['cor']
                result['sub'] = result['sub'] + self.data[token]['sub']
                result['ins'] = result['ins'] + self.data[token]['ins']
                result['del'] = result['del'] + self.data[token]['del']
        return result

    def keys(self):
        return list(self.data.keys())


def width(string):
    return sum(1 + (unicodedata.east_asian_width(c) in "AFW") for c in string)


def default_cluster(word):
    unicode_names = [unicodedata.name(char) for char in word]
    for i in reversed(range(len(unicode_names))):
        if unicode_names[i].startswith('DIGIT'):  # 1
            unicode_names[i] = 'Number'  # 'DIGIT'
        elif (unicode_names[i].startswith('CJK UNIFIED IDEOGRAPH')
              or unicode_names[i].startswith('CJK COMPATIBILITY IDEOGRAPH')):
            # 明 / 郎
            unicode_names[i] = 'Mandarin'  # 'CJK IDEOGRAPH'
        elif (unicode_names[i].startswith('LATIN CAPITAL LETTER')
              or unicode_names[i].startswith('LATIN SMALL LETTER')):
            # A / a
            unicode_names[i] = 'English'  # 'LATIN LETTER'
        elif unicode_names[i].startswith('HIRAGANA LETTER'):  # は こ め
            unicode_names[i] = 'Japanese'  # 'GANA LETTER'
        elif (unicode_names[i].startswith('AMPERSAND')
              or unicode_names[i].startswith('APOSTROPHE')
              or unicode_names[i].startswith('COMMERCIAL AT')
              or unicode_names[i].startswith('DEGREE CELSIUS')
              or unicode_names[i].startswith('EQUALS SIGN')
              or unicode_names[i].startswith('FULL STOP')
              or unicode_names[i].startswith('HYPHEN-MINUS')
              or unicode_names[i].startswith('LOW LINE')
              or unicode_names[i].startswith('NUMBER SIGN')
              or unicode_names[i].startswith('PLUS SIGN')
              or unicode_names[i].startswith('SEMICOLON')):
            # & / ' / @ / ℃ / = / . / - / _ / # / + / ;
            del unicode_names[i]
        else:
            return 'Other'
    if len(unicode_names) == 0:
        return 'Other'
    if len(unicode_names) == 1:
        return unicode_names[0]
    for i in range(len(unicode_names) - 1):
        if unicode_names[i] != unicode_names[i + 1]:
            return 'Other'
    return unicode_names[0]


def usage():
    print(
        "compute_wer.py : compute word error rate (WER) and align recognition results and references."
    )
    print(
        "         usage : python compute_wer.py [--cs={0,1}] [--cluster=foo] [--ig=ignore_file] [--char={0,1}] [--v={0,1}] [--padding-symbol={space,underline}] test.ref test.hyp > test.wer"
    )


if __name__ == '__main__':
    if len(sys.argv) == 1:
        usage()
        sys.exit(0)
    calculator = Calculator()
    cluster_file = ''
    ignore_words = set()
    tochar = False
    verbose = 1
    padding_symbol = ' '
    case_sensitive = False
    max_words_per_line = sys.maxsize
    split = None
    while len(sys.argv) > 3:
        a = '--maxw='
        if sys.argv[1].startswith(a):
            b = sys.argv[1][len(a):]
            del sys.argv[1]
            max_words_per_line = int(b)
            continue
        a = '--rt='
        if sys.argv[1].startswith(a):
            b = sys.argv[1][len(a):].lower()
            del sys.argv[1]
            remove_tag = (b == 'true') or (b != '0')
            continue
        a = '--cs='
        if sys.argv[1].startswith(a):
            b = sys.argv[1][len(a):].lower()
            del sys.argv[1]
            case_sensitive = (b == 'true') or (b != '0')
            continue
        a = '--cluster='
        if sys.argv[1].startswith(a):
            cluster_file = sys.argv[1][len(a):]
            del sys.argv[1]
            continue
        a = '--splitfile='
        if sys.argv[1].startswith(a):
            split_file = sys.argv[1][len(a):]
            del sys.argv[1]
            split = dict()
            with codecs.open(split_file, 'r', 'utf-8') as fh:
                for line in fh:  # line in unicode
                    words = line.strip().split()
                    if len(words) >= 2:
                        split[words[0]] = words[1:]
            continue
        a = '--ig='
        if sys.argv[1].startswith(a):
            ignore_file = sys.argv[1][len(a):]
            del sys.argv[1]
            with codecs.open(ignore_file, 'r', 'utf-8') as fh:
                for line in fh:  # line in unicode
                    line = line.strip()
                    if len(line) > 0:
                        ignore_words.add(line)
            continue
        a = '--char='
        if sys.argv[1].startswith(a):
            b = sys.argv[1][len(a):].lower()
            del sys.argv[1]
            tochar = (b == 'true') or (b != '0')
            continue
        a = '--v='
        if sys.argv[1].startswith(a):
            b = sys.argv[1][len(a):].lower()
            del sys.argv[1]
            verbose = 0
            try:
                verbose = int(b)
            except:
                if b == 'true' or b != '0':
                    verbose = 1
            continue
        a = '--padding-symbol='
        if sys.argv[1].startswith(a):
            b = sys.argv[1][len(a):].lower()
            del sys.argv[1]
            if b == 'space':
                padding_symbol = ' '
            elif b == 'underline':
                padding_symbol = '_'
            continue
        if True or sys.argv[1].startswith('-'):
            #ignore invalid switch
            del sys.argv[1]
            continue

    if not case_sensitive:
        ig = set([w.upper() for w in ignore_words])
        ignore_words = ig

    default_clusters = {}
    default_words = {}

    ref_file = sys.argv[1]
    hyp_file = sys.argv[2]
    rec_set = {}
    if split and not case_sensitive:
        newsplit = dict()
        for w in split:
            words = split[w]
            for i in range(len(words)):
                words[i] = words[i].upper()
            newsplit[w.upper()] = words
        split = newsplit

    with codecs.open(hyp_file, 'r', 'utf-8') as fh:
        for line in fh:
            if tochar:
                array = characterize(line)
            else:
                array = line.strip().split()
            if len(array) == 0: continue
            fid = array[0]
            rec_set[fid] = normalize(array[1:], ignore_words, case_sensitive,
                                     split)

    # compute error rate on the interaction of reference file and hyp file
    for line in open(ref_file, 'r', encoding='utf-8'):
        if tochar:
            array = characterize(line)
        else:
            array = line.rstrip('\n').split()
        if len(array) == 0: continue
        fid = array[0]
        if fid not in rec_set:
            continue
        lab = normalize(array[1:], ignore_words, case_sensitive, split)
        rec = rec_set[fid]
        if verbose:
            print('\nutt: %s' % fid)

        for word in rec + lab:
            if word not in default_words:
                default_cluster_name = default_cluster(word)
                if default_cluster_name not in default_clusters:
                    default_clusters[default_cluster_name] = {}
                if word not in default_clusters[default_cluster_name]:
                    default_clusters[default_cluster_name][word] = 1
                default_words[word] = default_cluster_name

        result = calculator.calculate(lab, rec)
        if verbose:
            if result['all'] != 0:
                wer = float(result['ins'] + result['sub'] +
                            result['del']) * 100.0 / result['all']
            else:
                wer = 0.0
            print('WER: %4.2f %%' % wer, end=' ')
            print('N=%d C=%d S=%d D=%d I=%d' %
                  (result['all'], result['cor'], result['sub'], result['del'],
                   result['ins']))
            space = {}
            space['lab'] = []
            space['rec'] = []
            for idx in range(len(result['lab'])):
                len_lab = width(result['lab'][idx])
                len_rec = width(result['rec'][idx])
                length = max(len_lab, len_rec)
                space['lab'].append(length - len_lab)
                space['rec'].append(length - len_rec)
            upper_lab = len(result['lab'])
            upper_rec = len(result['rec'])
            lab1, rec1 = 0, 0
            while lab1 < upper_lab or rec1 < upper_rec:
                if verbose > 1:
                    print('lab(%s):' % fid.encode('utf-8'), end=' ')
                else:
                    print('lab:', end=' ')
                lab2 = min(upper_lab, lab1 + max_words_per_line)
                for idx in range(lab1, lab2):
                    token = result['lab'][idx]
                    print('{token}'.format(token=token), end='')
                    for n in range(space['lab'][idx]):
                        print(padding_symbol, end='')
                    print(' ', end='')
                print()
                if verbose > 1:
                    print('rec(%s):' % fid.encode('utf-8'), end=' ')
                else:
                    print('rec:', end=' ')
                rec2 = min(upper_rec, rec1 + max_words_per_line)
                for idx in range(rec1, rec2):
                    token = result['rec'][idx]
                    print('{token}'.format(token=token), end='')
                    for n in range(space['rec'][idx]):
                        print(padding_symbol, end='')
                    print(' ', end='')
                print('\n', end='\n')
                lab1 = lab2
                rec1 = rec2

    if verbose:
        print(
            '==========================================================================='
        )
        print()

    result = calculator.overall()
    if result['all'] != 0:
        wer = float(result['ins'] + result['sub'] +
                    result['del']) * 100.0 / result['all']
    else:
        wer = 0.0
    print('Overall -> %4.2f %%' % wer, end=' ')
    print('N=%d C=%d S=%d D=%d I=%d' %
          (result['all'], result['cor'], result['sub'], result['del'],
           result['ins']))
    if not verbose:
        print()

    if verbose:
        for cluster_id in default_clusters:
            result = calculator.cluster(
                [k for k in default_clusters[cluster_id]])
            if result['all'] != 0:
                wer = float(result['ins'] + result['sub'] +
                            result['del']) * 100.0 / result['all']
            else:
                wer = 0.0
            print('%s -> %4.2f %%' % (cluster_id, wer), end=' ')
            print('N=%d C=%d S=%d D=%d I=%d' %
                  (result['all'], result['cor'], result['sub'], result['del'],
                   result['ins']))
        if len(cluster_file) > 0:  # compute separated WERs for word clusters
            cluster_id = ''
            cluster = []
            for line in open(cluster_file, 'r', encoding='utf-8'):
                for token in line.decode('utf-8').rstrip('\n').split():
                    # end of cluster reached, like </Keyword>
                    if token[0:2] == '</' and token[len(token)-1] == '>' and \
                       token.lstrip('</').rstrip('>') == cluster_id :
                        result = calculator.cluster(cluster)
                        if result['all'] != 0:
                            wer = float(result['ins'] + result['sub'] +
                                        result['del']) * 100.0 / result['all']
                        else:
                            wer = 0.0
                        print('%s -> %4.2f %%' % (cluster_id, wer), end=' ')
                        print('N=%d C=%d S=%d D=%d I=%d' %
                              (result['all'], result['cor'], result['sub'],
                               result['del'], result['ins']))
                        cluster_id = ''
                        cluster = []
                    # begin of cluster reached, like <Keyword>
                    elif token[0] == '<' and token[len(token)-1] == '>' and \
                         cluster_id == '' :
                        cluster_id = token.lstrip('<').rstrip('>')
                        cluster = []
                    # general terms, like WEATHER / CAR / ...
                    else:
                        cluster.append(token)
        print()
        print(
            '==========================================================================='
        )



# Copyright (c) 2021 Nippon Telegraph and Telephone corporation (NTT).
# All rights reserved

import numpy as np
import torch
import torch.nn.functional as F
import torch.nn as nn
from torch.optim.lr_scheduler import _LRScheduler


class NoamScheduler(_LRScheduler):
    """ learning rate scheduler used in the transformer
    See https://arxiv.org/pdf/1706.03762.pdf
    lrate = d_model**(-0.5) * \
            min(step_num**(-0.5), step_num*warmup_steps**(-1.5))
    Scaling factor is implemented as in
        http://nlp.seas.harvard.edu/2018/04/03/attention.html#optimizer
    """

    def __init__(
            self, optimizer, d_model, warmup_steps, tot_step, scale,
            last_epoch=-1
            ):
        self.d_model = d_model
        self.warmup_steps = warmup_steps
        self.tot_step = tot_step
        self.scale = scale
        super(NoamScheduler, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        self.last_epoch = max(1, self.last_epoch)
        step_num = self.last_epoch
        val = self.scale * self.d_model ** (-0.5) * \
            min(step_num ** (-0.5), step_num * self.warmup_steps ** (-1.5))

        return [base_lr / base_lr * val for base_lr in self.base_lrs]


class MultiHeadSelfAttention(nn.Module):
    """ Multi head "self" attention layer
    """

    def __init__(self, n_units, h=8, dropout_rate=0.1):
        super(MultiHeadSelfAttention, self).__init__()
        self.linearQ = nn.Linear(n_units, n_units)
        self.linearK = nn.Linear(n_units, n_units)
        self.linearV = nn.Linear(n_units, n_units)
        self.linearO = nn.Linear(n_units, n_units)
        self.d_k = n_units // h
        self.h = h
        self.dropout = nn.Dropout(p=dropout_rate)
        # attention for plot
        self.att = None

    def forward(self, x, batch_size):
        # x: (BT, F)
        q = self.linearQ(x).reshape(batch_size, -1, self.h, self.d_k)
        k = self.linearK(x).reshape(batch_size, -1, self.h, self.d_k)
        v = self.linearV(x).reshape(batch_size, -1, self.h, self.d_k)

        scores = torch.matmul(
                 q.transpose(1, 2), k.permute(0, 2, 3, 1)) / np.sqrt(self.d_k)
        # scores: (B, h, T, T) = (B, h, T, d_k) x (B, h, d_k, T)
        self.att = F.softmax(scores, dim=3)
        p_att = self.dropout(self.att)
        x = torch.matmul(p_att, v.transpose(1, 2))
        x = x.transpose(1, 2).reshape(-1, self.h * self.d_k)

        return self.linearO(x)


class PositionwiseFeedForward(nn.Module):
    """ Positionwise feed-forward layer
    """

    def __init__(self, n_units, d_units, dropout_rate):
        super(PositionwiseFeedForward, self).__init__()
        self.linear1 = nn.Linear(n_units, d_units)
        self.linear2 = nn.Linear(d_units, n_units)
        self.dropout = nn.Dropout(p=dropout_rate)

    def forward(self, x):
        return self.linear2(self.dropout(F.relu(self.linear1(x))))


class PositionalEncoding(nn.Module):
    """ Positional encoding function
    """

    def __init__(self, n_units, dropout_rate, max_len):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout_rate)
        positions = np.arange(0, max_len, dtype='f')[:, None]
        dens = np.exp(
            np.arange(0, n_units, 2, dtype='f') * -(np.log(10000.) / n_units))
        self.enc = np.zeros((max_len, n_units), dtype='f')
        self.enc[:, ::2] = np.sin(positions * dens)
        self.enc[:, 1::2] = np.cos(positions * dens)
        self.scale = np.sqrt(n_units)

    def forward(self, x):
        x = x * self.scale + self.xp.array(self.enc[:, :x.shape[1]])
        return self.dropout(x)


class TransformerEncoder(nn.Module):
    def __init__(self, idim, n_layers, n_units,
                 e_units=2048, h=8, dropout_rate=0.1):
        super(TransformerEncoder, self).__init__()
        self.linear_in = nn.Linear(idim, n_units)
        # self.lnorm_in = nn.LayerNorm(n_units)
        self.pos_enc = PositionalEncoding(n_units, dropout_rate, 5000)
        self.n_layers = n_layers
        self.dropout = nn.Dropout(p=dropout_rate)
        for i in range(n_layers):
            setattr(self, '{}{:d}'.format("lnorm1_", i),
                    nn.LayerNorm(n_units))
            setattr(self, '{}{:d}'.format("self_att_", i),
                    MultiHeadSelfAttention(n_units, h, dropout_rate))
            setattr(self, '{}{:d}'.format("lnorm2_", i),
                    nn.LayerNorm(n_units))
            setattr(self, '{}{:d}'.format("ff_", i),
                    PositionwiseFeedForward(n_units, e_units, dropout_rate))
        self.lnorm_out = nn.LayerNorm(n_units)

    def forward(self, x):
        # x: (B, T, F) ... batch, time, (mel)freq
        BT_size = x.shape[0] * x.shape[1]
        # e: (BT, F)
        e = self.linear_in(x.reshape(BT_size, -1))
        # Encoder stack
        for i in range(self.n_layers):
            # layer normalization
            e = getattr(self, '{}{:d}'.format("lnorm1_", i))(e)
            # self-attention
            s = getattr(self, '{}{:d}'.format("self_att_", i))(e, x.shape[0])
            # residual
            e = e + self.dropout(s)
            # layer normalization
            e = getattr(self, '{}{:d}'.format("lnorm2_", i))(e)
            # positionwise feed-forward
            s = getattr(self, '{}{:d}'.format("ff_", i))(e)
            # residual
            e = e + self.dropout(s)
        # final layer normalization
        # output: (BT, F)
        return self.lnorm_out(e)



from dataclasses import dataclass
import json
import torch
import threading
import queue
import subprocess
import concurrent.futures

import numpy as np
import torchaudio.compliance.kaldi as k
import soundfile as sf

class PCMQueue:
    def __init__(self):
        """
        Initialize the PCMQueue with an empty queue, an empty buffer, and a lock for thread safety.
        
        Parameters:
        - None
        
        Returns:
        - None
        """
        self.queue = queue.Queue()
        self.buffer = np.array([], dtype=np.float32)
        self.lock = threading.Lock()

    def put(self, items):
        """
        Add items to the buffer in a thread-safe manner.
        
        Parameters:
        - items (list or array-like): The items to be added to the buffer, a numpy array of dtype np.float32.
        
        Returns:
        - None
        """
        with self.lock:
            self.buffer = np.concatenate((self.buffer, np.array(items, dtype=np.float32)))

    def get(self, length):
        """
        Retrieve a specified number of elements from the buffer in a thread-safe manner.
        
        Parameters:
        - length (int): The number of elements to retrieve from the buffer.
        
        Returns:
        - numpy.ndarray or None: A numpy array containing the requested number of elements if available, otherwise None.
        """
        with self.lock:
            if len(self.buffer) < length:
                return None
            result = self.buffer[:length]
            self.buffer = self.buffer[length:]
            return result
    
    def clear(self):
        """
        Clear the buffer in a thread-safe manner.
        
        Parameters:
        - None
        
        Returns:
        - None
        """
        with self.lock:
            self.buffer = np.array([], dtype=np.float32)

    def has_enough_data(self, length):
        """
        Check if the buffer contains enough data to fulfill a request of a specified length.
        
        Parameters:
        - length (int): The number of elements required.
        
        Returns:
        - bool: True if the buffer contains enough data, False otherwise.
        """
        with self.lock:
            return len(self.buffer) >= length

class ThreadSafeQueue:
    def __init__(self):
        """
        Initialize the ThreadSafeQueue with an empty queue and a lock for thread safety.
        
        Parameters:
        - None
        
        Returns:
        - None
        """
        self._queue = queue.Queue()
        self._lock = threading.Lock()

    def put(self, item):
        """
        Add an item to the queue in a thread-safe manner.
        
        Parameters:
        - item (any): The item to be added to the queue.
        
        Returns:
        - None
        """
        with self._lock:
            self._queue.put(item)

    def get(self):
        """
        Retrieve an item from the queue in a thread-safe manner.
        
        Parameters:
        - None
        
        Returns:
        - any or None: The retrieved item if the queue is not empty, otherwise None.
        """
        with self._lock:
            if not self._queue.empty():
                return self._queue.get()
            else:
                return None

    def clear(self):
        """
        Clear the queue in a thread-safe manner.
        
        Parameters:
        - None
        
        Returns:
        - None
        """
        with self._lock:
            while not self._queue.empty():
                self._queue.get()

    def is_empty(self):
        """
        Check if the queue is empty in a thread-safe manner.
        
        Parameters:
        - None
        
        Returns:
        - bool: True if the queue is empty, False otherwise.
        """
        with self._lock:
            return self._queue.empty()

    def size(self):
        """
        Get the current size of the queue in a thread-safe manner.
        
        Parameters:
        - None
        
        Returns:
        - int: The number of items currently in the queue.
        """
        with self._lock:
            return self._queue.qsize()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import torch
import math
from dataclasses import dataclass, field

from fairseq import pdb
from fairseq import utils, metrics
from fairseq.criterions import FairseqCriterion, register_criterion
from fairseq.criterions.wav2vec_criterion import Wav2vecCriterion, Wav2VecCriterionConfig
from fairseq.criterions.ctc import CtcCriterion, CtcCriterionConfig
from fairseq.logging.meters import safe_round


@dataclass
class UnispeechCriterionConfig(Wav2VecCriterionConfig, CtcCriterionConfig):
    mtlalpha: float = field(
        default=0.5, metadata={"help": "loss weight for multitask learning"}
    )
    

@register_criterion('unispeech_criterion', dataclass=UnispeechCriterionConfig)
class UnispeechCriterion(FairseqCriterion):

    def __init__(self, cfg:UnispeechCriterionConfig, task):
        super().__init__(task)
        self.mtlalpha = cfg.mtlalpha
        self.w2v_criterion = Wav2vecCriterion(task, cfg.infonce, cfg.loss_weights, cfg.log_keys)
        if self.mtlalpha > 0:
            self.ctc_criterion = CtcCriterion(cfg, task)


    def forward(self, model, sample, reduce=True):
        net_output = model(**sample["net_input"])
     
        if self.mtlalpha > 0.0:
            ctc_loss, ctc_sample_size, ctc_logging_output = self.ctc_criterion.get_loss(model, sample, net_output, reduce)
        else:
            ctc_loss = 0
            ctc_sample_size = 0
            ctc_logging_output = {}

        infonce_loss, infonce_sample_size, infonce_logging_output = self.w2v_criterion.get_loss(model.w2v_encoder.w2v_model, sample, net_output['contrastive_res'], reduce)
        loss = self.mtlalpha * ctc_loss + (1.0 - self.mtlalpha) * infonce_loss
        sample_size = infonce_sample_size
        logging_output = {'loss': loss, 'ntokens': ctc_logging_output['ntokens'], 'nsentences': ctc_logging_output['nsentences'],
                           'ctc': ctc_logging_output, 'infonce': infonce_logging_output}

        return loss, sample_size, logging_output

    @staticmethod
    def logging_outputs_can_be_summed() -> bool:
        return False

    @staticmethod
    def reduce_metrics(logging_outputs) -> None:
        loss_sum = utils.item(sum(log.get("loss", 0) for log in logging_outputs))

        ctc_loss_sum = utils.item(sum(log['ctc'].get('loss', 0) for log in logging_outputs))
        ctc_sample_size = utils.item(sum(log['ctc'].get('sample_size', 0) for log in logging_outputs))
        ctc_ntokens = utils.item(sum(log['ctc'].get('ntokens', 0) for log in logging_outputs))
        ctc_nsentences = utils.item(sum(log['ctc'].get('nsentences', 0) for log in logging_outputs))

        ctras_loss_sum = utils.item(sum(log['infonce'].get('loss', 0) for log in logging_outputs)) 
        ctras_sample_size = utils.item(sum(log['infonce'].get('sample_size', 0) for log in logging_outputs))
        ctras_ntokens = utils.item(sum(log['infonce'].get('ntokens', 0) for log in logging_outputs))
        ctras_nsentences = utils.item(sum(log['infonce'].get('nsentences', 0) for log in logging_outputs))

        metrics.log_scalar(
            "loss", loss_sum, 1, round=3)
        metrics.log_scalar(
            "ctc_loss", ctc_loss_sum / ctc_sample_size / math.log(2), ctc_sample_size, round=3
        )
        metrics.log_scalar(
            "contrastive_loss", ctras_loss_sum / ctras_sample_size / math.log(2), ctras_sample_size, round=3
        )
        if ctc_sample_size != ctc_ntokens:
            metrics.log_scalar(
                "nll_loss", ctc_loss_sum / ctc_ntokens / math.log(2), ctc_ntokens, round=3
            )
        c_errors = sum(log['ctc'].get("c_errors", 0) for log in logging_outputs)
        metrics.log_scalar("_c_errors", c_errors)
        c_total = sum(log['ctc'].get("c_total", 0) for log in logging_outputs)
        metrics.log_scalar("_c_total", c_total)
        w_errors = sum(log['ctc'].get("w_errors", 0) for log in logging_outputs)
        metrics.log_scalar("_w_errors", w_errors)
        wv_errors = sum(log['ctc'].get("wv_errors", 0) for log in logging_outputs)
        metrics.log_scalar("_wv_errors", wv_errors)
        w_total = sum(log['ctc'].get("w_total", 0) for log in logging_outputs)
        metrics.log_scalar("_w_total", w_total)

        if c_total > 0:
             metrics.log_derived(
                "uer",
                lambda meters: safe_round(meters["_c_errors"].sum * 100.0 / meters["_c_total"].sum, 3)
                if meters["_c_total"].sum > 0
                else float("nan"),
            )
        if w_total > 0:
            metrics.log_derived(
                "wer",
                lambda meters: safe_round(meters["_w_errors"].sum * 100.0 / meters["_w_total"].sum, 3)
                if meters["_w_total"].sum > 0
                else float("nan"),
            )
            metrics.log_derived(
                "raw_wer",
                lambda meters: safe_round(meters["_wv_errors"].sum * 100.0 / meters["_w_total"].sum, 3)
                if meters["_w_total"].sum > 0
                else float("nan"),
            )


        metrics.log_scalar("nsentences", ctras_nsentences)
        metrics.log_scalar("ctc_sample_size", ctc_sample_size)
        metrics.log_scalar("contrastive_sample_size", ctras_sample_size)

                
        correct = sum(log['infonce'].get("correct", 0) for log in logging_outputs)
        metrics.log_scalar("_correct", correct)
  
        total = sum(log['infonce'].get("count", 0) for log in logging_outputs)
        metrics.log_scalar("_total", total)


        if total > 0:
            metrics.log_derived(
                "accuracy",
                lambda meters: safe_round(meters["_correct"].sum / meters["_total"].sum, 5)
                if meters["_total"].sum > 0
               else float("nan"),
            )

        builtin_keys = {'loss', 'ntokens', 'nsentences', 'sample_size', 'correct', 'count'}
        for k in logging_outputs[0]['infonce']:
            if k not in builtin_keys:
                val = sum(log['infonce'].get(k, 0) for log in logging_outputs) / len(logging_outputs)
                if k.startswith('loss'):
                    metrics.log_scalar(k, val / ctras_sample_size / math.log(2), ctras_sample_size)
                else:
                    metrics.log_scalar(k, val, round=3)    




import os
import sys

sys.path.append(os.getcwd())

import json
from importlib.resources import files
from pathlib import Path
from tqdm import tqdm
import soundfile as sf
from datasets.arrow_writer import ArrowWriter


def main():
    result = []
    duration_list = []
    text_vocab_set = set()

    with open(meta_info, "r") as f:
        lines = f.readlines()
        for line in tqdm(lines):
            uttr, text, norm_text = line.split("|")
            norm_text = norm_text.strip()
            wav_path = Path(dataset_dir) / "wavs" / f"{uttr}.wav"
            duration = sf.info(wav_path).duration
            if duration < 0.4 or duration > 30:
                continue
            result.append({"audio_path": str(wav_path), "text": norm_text, "duration": duration})
            duration_list.append(duration)
            text_vocab_set.update(list(norm_text))

    # save preprocessed dataset to disk
    if not os.path.exists(f"{save_dir}"):
        os.makedirs(f"{save_dir}")
    print(f"\nSaving to {save_dir} ...")

    with ArrowWriter(path=f"{save_dir}/raw.arrow") as writer:
        for line in tqdm(result, desc="Writing to raw.arrow ..."):
            writer.write(line)

    # dup a json separately saving duration in case for DynamicBatchSampler ease
    with open(f"{save_dir}/duration.json", "w", encoding="utf-8") as f:
        json.dump({"duration": duration_list}, f, ensure_ascii=False)

    # vocab map, i.e. tokenizer
    # add alphabets and symbols (optional, if plan to ft on de/fr etc.)
    with open(f"{save_dir}/vocab.txt", "w") as f:
        for vocab in sorted(text_vocab_set):
            f.write(vocab + "\n")

    print(f"\nFor {dataset_name}, sample count: {len(result)}")
    print(f"For {dataset_name}, vocab size is: {len(text_vocab_set)}")
    print(f"For {dataset_name}, total {sum(duration_list)/3600:.2f} hours")


if __name__ == "__main__":
    tokenizer = "char"  # "pinyin" | "char"

    dataset_dir = "<SOME_PATH>/LJSpeech-1.1"
    dataset_name = f"LJSpeech_{tokenizer}"
    meta_info = os.path.join(dataset_dir, "metadata.csv")
    save_dir = str(files("f5_tts").joinpath("../../")) + f"/data/{dataset_name}"
    print(f"\nPrepare for {dataset_name}, will save to {save_dir}\n")

    main()



# Copyright (c) 2024 Alibaba Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import torchaudio
import time
import logging
import argparse
from inspiremusic.cli.inspiremusic import InspireMusic
from inspiremusic.utils.file_utils import logging
import torch
from inspiremusic.utils.audio_utils import trim_audio, fade_out, process_audio

def set_env_variables():
    os.environ['PYTHONIOENCODING'] = 'UTF-8'
    os.environ['TOKENIZERS_PARALLELISM'] = 'False'
    current_working_dir = os.getcwd()
    main_root = os.path.realpath(os.path.join(current_working_dir, '../../'))
    bin_dir = os.path.join(main_root, 'inspiremusic')
    third_party_matcha_tts_path = os.path.join(main_root, 'third_party', 'Matcha-TTS')
    python_path = f"{main_root}:{bin_dir}:{third_party_matcha_tts_path}:{os.environ.get('PYTHONPATH', '')}"
    os.environ['PYTHONPATH'] = python_path
    sys.path.extend([main_root, third_party_matcha_tts_path])

class InspireMusicUnified:
    def __init__(self,
                 model_name: str = "InspireMusic-1.5B-Long",
                 model_dir: str = None,
                 min_generate_audio_seconds: float = 10.0,
                 max_generate_audio_seconds: float = 30.0,
                 sample_rate: int = 24000,
                 output_sample_rate: int = 48000,
                 load_jit: bool = True,
                 load_onnx: bool = False,
                 dtype: str = "fp16",
                 fast: bool = False,
                 fp16: bool = True,
                 gpu: int = 0,
                 result_dir: str = None,
                 hub="modelscope"):
        os.environ['CUDA_VISIBLE_DEVICES'] = str(gpu)

        # Set model_dir or default to downloading if it doesn't exist
        if model_dir is None:
             model_dir = f"../../pretrained_models/{model_name}"

        if not os.path.isfile(f"{model_dir}/llm.pt"):
            if hub == "modelscope":
                from modelscope import snapshot_download
                if model_name == "InspireMusic-Base":
                    snapshot_download(f"iic/InspireMusic", local_dir=model_dir)
                else:
                    snapshot_download(f"iic/{model_name}", local_dir=model_dir)

        self.model_dir = model_dir

        self.sample_rate = sample_rate
        self.output_sample_rate = 24000 if fast else output_sample_rate
        self.result_dir = result_dir or f"exp/{model_name}"
        os.makedirs(self.result_dir, exist_ok=True)

        self.min_generate_audio_seconds = min_generate_audio_seconds
        self.max_generate_audio_seconds = max_generate_audio_seconds
        self.min_generate_audio_length = int(self.output_sample_rate * self.min_generate_audio_seconds)
        self.max_generate_audio_length = int(self.output_sample_rate * self.max_generate_audio_seconds)
        assert self.min_generate_audio_seconds <= self.max_generate_audio_seconds, "Min audio seconds must be less than or equal to max audio seconds"

        use_cuda = gpu >= 0 and torch.cuda.is_available()
        self.device = torch.device('cuda' if use_cuda else 'cpu')
        self.model = InspireMusic(self.model_dir, load_jit=load_jit, load_onnx=load_onnx, dtype=dtype, fast=fast, fp16=fp16)

        logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

    @torch.inference_mode()
    def inference(self,
                  task: str = 'text-to-music',
                  text: str = None,
                  audio_prompt: str = None, # audio prompt file path
                  instruct: str = None,
                  chorus: str = "verse",
                  time_start: float = 0.0,
                  time_end: float = 30.0,
                  output_fn: str = "output_audio",
                  max_audio_prompt_length: float = 5.0,
                  fade_out_duration: float = 1.0,
                  output_format: str = "wav",
                  fade_out_mode: bool = True,
                  trim: bool = False,
                  ):

        with torch.no_grad():
            text_prompt = f"<|{time_start}|><|{chorus}|><|{text}|><|{time_end}|>"
            chorus_dict = {"random": torch.randint(1, 5, (1,)).item(), "intro" : 0, "verse": 1, "chorus": 2, "outro": 4}
            chorus = chorus_dict.get(chorus, 1)
            chorus = torch.tensor([chorus], dtype=torch.int).to(self.device)

            time_start_tensor = torch.tensor([time_start], dtype=torch.float64).to(self.device)
            time_end_tensor = torch.tensor([time_end], dtype=torch.float64).to(self.device)

            music_fn = os.path.join(self.result_dir, f'{output_fn}.{output_format}')

            bench_start = time.time()

            if task == 'text-to-music':
                model_input = {
                    "text"           : text,
                    "audio_prompt"   : audio_prompt,
                    "time_start"     : time_start_tensor,
                    "time_end"       : time_end_tensor,
                    "chorus"         : chorus,
                    "task"           : task,
                    "stream"         : False,
                    "duration_to_gen": self.max_generate_audio_seconds,
                    "sr"             : self.sample_rate
                }
            elif task == 'continuation':
                if audio_prompt is not None:
                    audio, _ = process_audio(audio_prompt, self.sample_rate)
                    if audio.size(1) < self.sample_rate:
                        logging.warning("Warning: Input prompt audio length is shorter than 1s. Please provide an appropriate length audio prompt and try again.")
                        audio = None
                    else:
                        max_audio_prompt_length_samples = int(max_audio_prompt_length * self.sample_rate)
                        audio = audio[:, :max_audio_prompt_length_samples]  # Trimming prompt audio

                model_input = {
                    "text"           : text,
                    "audio_prompt"   : audio,
                    "time_start"     : time_start_tensor,
                    "time_end"       : time_end_tensor,
                    "chorus"         : chorus,
                    "task"           : task,
                    "stream"         : False,
                    "duration_to_gen": self.max_generate_audio_seconds,
                    "sr"             : self.sample_rate
                }

            music_audios = []
            for model_output in self.model.cli_inference(**model_input):
                music_audios.append(model_output['music_audio'])

            bench_end = time.time()

            if trim:
                music_audio = trim_audio(music_audios[0],
                                         sample_rate=self.output_sample_rate,
                                         threshold=0.05,
                                         min_silence_duration=0.8)
            else:
                music_audio = music_audios[0]

            if music_audio.shape[0] != 0:
                if music_audio.shape[1] > self.max_generate_audio_length:
                    music_audio = music_audio[:, :self.max_generate_audio_length]

                if music_audio.shape[1] >= self.min_generate_audio_length:
                    try:
                        if fade_out_mode:
                            music_audio = fade_out(music_audio, self.output_sample_rate, fade_out_duration)

                        music_audio = music_audio.repeat(2, 1)

                        if output_format in ["wav", "flac"]:
                            torchaudio.save(music_fn, music_audio,
                                            sample_rate=self.output_sample_rate,
                                            encoding="PCM_S",
                                            bits_per_sample=24)
                        elif output_format in ["mp3", "m4a"]:
                            torchaudio.backend.sox_io_backend.save(
                                filepath=music_fn, src=music_audio,
                                sample_rate=self.output_sample_rate,
                                format=output_format)
                        else:
                            logging.info("Format is not supported. Please choose from wav, mp3, m4a, flac.")

                    except Exception as e:
                        logging.error(f"Error saving file: {e}")
                        raise

                audio_duration = music_audio.shape[1] / self.output_sample_rate
                rtf = (bench_end - bench_start) / audio_duration
                logging.info(f"Processing time: {int(bench_end - bench_start)}s, audio length: {int(audio_duration)}s, rtf: {rtf}, text prompt: {text_prompt}")

            else:
                logging.error(f"Generated audio length is shorter than minimum required audio length.")
        if music_fn:
            if os.path.exists(music_fn):
                logging.info(f"Generated audio file {music_fn} is saved.")
                return music_fn
            else:
                logging.error(f"{music_fn} does not exist.")

def get_args():
    parser = argparse.ArgumentParser(description='Run inference with your model')
    parser.add_argument('-m', '--model_name', default="InspireMusic-1.5B-Long",
                        help='Model name')
    
    parser.add_argument('-d', '--model_dir',
                        help='Model folder path')
    
    parser.add_argument('-t', '--text', default="Experience soothing and sensual instrumental jazz with a touch of Bossa Nova, perfect for a relaxing restaurant or spa ambiance.",
                        help='Prompt text')
    
    parser.add_argument('-a', '--audio_prompt', default=None, 
                        help='Prompt audio')
    
    parser.add_argument('-c', '--chorus', default="intro", 
                        help='Chorus tag generation mode (e.g., random, verse, chorus, intro, outro)')
    
    parser.add_argument('-f', '--fast', type=bool, default=False, 
                        help='Enable fast inference mode (without flow matching)')

    parser.add_argument('-g', '--gpu', type=int, default=0, 
                        help='GPU ID for this rank, -1 for CPU')
    
    parser.add_argument('--task', default='text-to-music', choices=['text-to-music', 'continuation', 'reconstruct', 'super_resolution'],
                        help='Inference task type: text-to-music, continuation, reconstruct, super_resolution')
    
    parser.add_argument('-r', '--result_dir', default="exp/inspiremusic", 
                        help='Directory to save generated audio')
    
    parser.add_argument('-o', '--output_fn', default="output_audio", 
                        help='Output file name')
    
    parser.add_argument('--format', type=str, default="wav", choices=["wav", "mp3", "m4a", "flac"],
                        help='Format of output audio')
    
    parser.add_argument('--sample_rate', type=int, default=24000, 
                        help='Sampling rate of input audio')
    
    parser.add_argument('--output_sample_rate', type=int, default=48000, choices=[24000, 48000],
                        help='Sampling rate of generated output audio')
    
    parser.add_argument('-s', '--time_start', type=float, default=0.0, 
                        help='Start time in seconds')
    
    parser.add_argument('-e', '--time_end', type=float, default=30.0, 
                        help='End time in seconds')
    
    parser.add_argument('--max_audio_prompt_length', type=float, default=5.0, 
                        help='Maximum audio prompt length in seconds')
    
    parser.add_argument('--min_generate_audio_seconds', type=float, default=10.0, 
                        help='Minimum generated audio length in seconds')
    
    parser.add_argument('--max_generate_audio_seconds', type=float, default=30.0, 
                        help='Maximum generated audio length in seconds')

    parser.add_argument('--fp16', type=bool, default=True, 
                        help='Inference with fp16 model')
    
    parser.add_argument('--fade_out', type=bool, default=True, 
                        help='Apply fade out effect to generated audio')
    
    parser.add_argument('--fade_out_duration', type=float, default=1.0, 
                        help='Fade out duration in seconds')
    
    parser.add_argument('--trim', type=bool, default=False, 
                        help='Trim the silence ending of generated audio')    

    args = parser.parse_args()

    if not args.model_dir:
        args.model_dir = os.path.join("../../pretrained_models", args.model_name)

    print(args)
    return args

def main():
    set_env_variables()
    args = get_args()
    model = InspireMusicUnified(model_name = args.model_name,
                 model_dir = args.model_dir,
                 min_generate_audio_seconds = args.min_generate_audio_seconds,
                 max_generate_audio_seconds = args.max_generate_audio_seconds,
                 sample_rate = args.sample_rate,
                 output_sample_rate = args.output_sample_rate,
                 load_jit = True,
                 load_onnx = False,
                 dtype="fp16",
                 fast = args.fast,
                 fp16 = args.fp16,
                 gpu = args.gpu,
                 result_dir = args.result_dir)

    model.inference(task = args.task,
                text = args.text,
                audio_prompt = args.audio_prompt,
                chorus = args.chorus,
                time_start = args.time_start,
                time_end = args.time_end,
                output_fn = args.output_fn,
                max_audio_prompt_length = args.max_audio_prompt_length,
                fade_out_duration = args.fade_out_duration,
                output_format = args.format,
                fade_out_mode = args.fade_out,
                trim = args.trim)

if __name__ == "__main__":
    main()